"""Tests GRPC Wrapper class."""
from datetime import datetime
import logging
import os
from pathlib import Path
import re
import tempfile
import time
from typing import Final, Optional, cast
from unittest.mock import (
    ANY,
    AsyncMock,
    Mock,
    NonCallableMock,
    PropertyMock,
    call,
    create_autospec,
)

from _pytest.logging import LogCaptureFixture
from grpc import RpcError, StatusCode
import msgpack
import numpy as np
import pandas as pd
import pytest
from pytest import fixture, raises
from pytest_mock import MockerFixture
from requests import HTTPError, RequestException

from bitfount.federated.encryption import _AESEncryption, _RSAEncryption
from bitfount.federated.exceptions import DecryptError
from bitfount.federated.transport import utils
from bitfount.federated.transport.config import MessageServiceConfig
from bitfount.federated.transport.exceptions import BitfountMessageServiceError
from bitfount.federated.transport.message_service import (
    _MAX_STORAGE_SIZE_BYTES,
    _MAX_STORAGE_SIZE_MEGABYTES,
    _SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES,
    _AutoRetryMessageServiceStub,
    _BitfountMessage,
    _BitfountMessageType,
    _DecryptedBitfountMessage,
    _LargeObjectRequestHandler,
    _MessageEncryption,
    _MessageService,
    msgpackext_decode,
    msgpackext_encode,
)
from bitfount.federated.transport.protos.messages_pb2 import (
    BitfountMessage as GrpcBitfountMessage,
)
from bitfount.federated.transport.protos.messages_pb2 import (
    BitfountTask,
    BitfountTasks,
    BlobStorageData,
    CommunicationDetails,
    PodData,
    SuccessResponse,
)
from bitfount.federated.transport.protos.messages_pb2 import Acknowledgement
from bitfount.hub.authentication_flow import BitfountSession
from bitfount.hub.types import APIKeys
from bitfount.types import _S3PresignedPOSTFields, _S3PresignedPOSTURL, _S3PresignedURL
from bitfount.utils import _get_mb_from_bytes
from tests.utils import PytestRequest
from tests.utils.fixtures import RpcErrorMaker
from tests.utils.helper import create_dataset, get_debug_logs, unit_test
from tests.utils.mocks import create_dataclass_mock

_GRPC_ERROR_CODES: Final = {sc for sc in StatusCode}
ACCESS_TOKEN: str = "someFakeAccessToken"


@unit_test
class TestMessageService:
    """Test MessageService (GRPC wrapper)."""

    @fixture
    def username(self) -> str:
        """Username."""
        return "theAuthenticatedUser"

    @fixture
    def task_id(self) -> str:
        """Task ID."""
        return "this-is-a-task-id"

    @fixture(params=(True, False), ids=("task_id_incl", "no_task_id"))
    def opt_task_id(self, request: PytestRequest, task_id: str) -> Optional[str]:
        """Returns a task ID or None, to cover both cases."""
        incl_task_id: bool = request.param
        if incl_task_id:
            return task_id
        else:
            return None

    @fixture
    def session(self, username: str) -> Mock:
        """Mocked session."""
        session: Mock = create_autospec(BitfountSession, instance=True)
        session.access_token = ACCESS_TOKEN
        session.username = username
        session.api_keys = None
        return session

    @fixture
    def grpc_stub(self) -> Mock:
        """Mocked GRPC Stub.

        Mocking this is a bit unpleasant due to the way it is constructed.
        """
        mock_stub = AsyncMock(
            spec_set=[
                "PodConnect",
                "SetupTaskMailboxes",
                "SendBitfountMessage",
                "GetBitfountMessage",
                "GetLargeObjectStorage",
                "AcknowledgeMessage",
            ]
        )

        # Need to explicitly set AsyncMocks on all the methods due to how
        # spec_set works
        mock_stub.PodConnect = AsyncMock()
        mock_stub.SetupTaskMailboxes = AsyncMock()
        mock_stub.SendBitfountMessage = AsyncMock()
        mock_stub.GetBitfountMessage = AsyncMock()
        mock_stub.GetLargeObjectStorage = AsyncMock()
        mock_stub.AcknowledgeMessage = AsyncMock()

        return mock_stub

    @fixture
    def mock_config_stub_property(self, grpc_stub: Mock) -> PropertyMock:
        """Property mock of the MessageServiceConfig.stub property."""
        # Need to use small helper function to get the async behaviour.
        async def _mock_stub_property() -> Mock:
            return grpc_stub

        return PropertyMock(side_effect=_mock_stub_property)

    @fixture
    def mock_ms_config(
        self, mock_config_stub_property: PropertyMock
    ) -> NonCallableMock:
        """Returns mock MessageServiceConfig."""
        mock_ms_config = create_dataclass_mock(MessageServiceConfig)

        # PropertyMock must be set on type(x), not x.
        type(mock_ms_config).stub = mock_config_stub_property

        # Explicitly set use_local_storage to False
        mock_ms_config.use_local_storage = False

        return mock_ms_config

    @fixture
    def message_service(
        self, mock_ms_config: NonCallableMock, session: Mock
    ) -> _MessageService:
        """The MessageService under test."""
        return _MessageService(session, mock_ms_config)

    @fixture
    def mock_upload_to_s3(self, mocker: MockerFixture) -> AsyncMock:
        """Mock out upload_data_to_s3 function in message_service.py."""
        mock_upload_to_s3 = mocker.patch(
            "bitfount.federated.transport.message_service._async_upload_data_to_s3",
            autospec=True,
        )
        return mock_upload_to_s3

    @fixture
    def mock_download_from_s3(self, mocker: MockerFixture) -> AsyncMock:
        """Mock out download_data_from_s3 function in message_service.py."""
        mock_download_from_s3 = mocker.patch(
            "bitfount.federated.transport.message_service._async_download_data_from_s3",
            autospec=True,
        )
        return mock_download_from_s3

    def test_metadata_property_with_api_keys(
        self, message_service: _MessageService
    ) -> None:
        """Test metadata property with API keys.

        Checks that only the first portion of the access key is included in the
        metadata.
        """
        message_service._session.api_keys = APIKeys(
            "accessKeyID1:accessKeyID2", "accessKey1:accessKey2"
        )
        assert message_service.metadata == [
            ("x-api-key-id", "accessKeyID1"),
            ("x-api-key", "accessKey1"),
        ]

    def test_metadata_property_without_api_keys(
        self, message_service: _MessageService, mocker: MockerFixture
    ) -> None:
        """Test metadata property without API keys simply returns the access token."""
        assert message_service.metadata == [("token", ACCESS_TOKEN)]

    def test_username_property_returns_username(
        self, message_service: _MessageService, username: str
    ) -> None:
        """Property returns username from session."""
        assert message_service.username == username

    async def test_stub_property(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mock_config_stub_property: PropertyMock,
    ) -> None:
        """Tests that the stub property/generates the stub from config."""
        # Check that the stub generated matches the output of the
        # MessageServiceConfig stub.
        stub = await message_service.stub
        assert isinstance(stub, _AutoRetryMessageServiceStub)
        assert stub._orig == grpc_stub

        # Check value is "cached"
        stub2 = await message_service.stub
        assert stub is stub2
        assert message_service._grpc.stub is not None

        # Check config stub only called once
        mock_config_stub_property.assert_called_once()

    async def test_connect_successful(
        self, grpc_stub: Mock, message_service: _MessageService, mocker: MockerFixture
    ) -> None:
        """Pod Connect successful."""
        platform = mocker.patch("bitfount.federated.transport.message_service.platform")
        psutil = mocker.patch("bitfount.federated.transport.message_service.psutil")
        get_gpu_metadata = mocker.patch(
            "bitfount.federated.transport.message_service.get_gpu_metadata"
        )
        platform.processor.return_value = "someProcessor"
        platform.system.return_value = "someOS"
        psutil.cpu_count.return_value = 50
        get_gpu_metadata.return_value = "someGPUName", 33

        memory_data = Mock()
        memory_data.total = 346346436
        psutil.virtual_memory.return_value = memory_data
        grpc_stub.PodConnect.return_value = SuccessResponse()
        pod_name = "somePodName"

        mailbox_id = await message_service.connect_pod(pod_name)

        assert mailbox_id == pod_name
        grpc_stub.PodConnect.assert_called_once_with(
            PodData(
                podName=pod_name,
                processor="someProcessor",
                podOS="someOS",
                cpuCount=50,
                gpuCount=33,
                gpuName="someGPUName",
                totalMemoryBytes=346346436,
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

    @pytest.mark.parametrize(
        argnames="status_code",
        argvalues=_GRPC_ERROR_CODES,
        ids=lambda sc: f"status_code={sc.name}",
    )
    async def test_connect_raises_error_on_failure(
        self,
        caplog: LogCaptureFixture,
        grpc_stub: Mock,
        message_service: _MessageService,
        remove_grpc_retry_backoff_sleep: AsyncMock,
        rpc_error_maker: RpcErrorMaker,
        status_code: StatusCode,
    ) -> None:
        """Pod Connect receives RpcError."""
        if status_code in utils._RETRY_STATUS_CODES:
            retries = True
            expected_num_calls = utils._DEFAULT_MAX_RETRIES + 1
        else:
            retries = False
            expected_num_calls = 1
        expected_num_retries = expected_num_calls - 1

        grpc_stub.PodConnect.side_effect = rpc_error_maker(
            "Pod connect failed", status_code
        )
        pod_name = "somePodName"

        with raises(
            BitfountMessageServiceError,
            match=re.escape(
                "Unable to connect pod to message service."
                " See debug logs for more details."
            ),
        ), caplog.at_level(logging.DEBUG):
            await message_service.connect_pod(pod_name)

        # Sanity check that PodConnect was called at all
        assert (
            grpc_stub.PodConnect.call_args_list
            == [
                call(
                    ANY,
                    metadata=[("token", ACCESS_TOKEN)],
                    timeout=utils._DEFAULT_TIMEOUT,
                )
            ]
            * expected_num_calls
        )

        # Check underlying exception was logged to DEBUG
        debug_logs = get_debug_logs(caplog, full_details=True)
        assert "RpcError: Pod connect failed" in debug_logs

        # Check retry messages in debug logs and sleep called if required
        if retries:
            assert remove_grpc_retry_backoff_sleep.await_count == expected_num_retries
            for i in range(1, expected_num_retries + 1):
                assert re.search(
                    rf"gRPC error occurred: .* Pod connect failed; .* \(attempt {i}\)",
                    debug_logs,
                )
        else:
            remove_grpc_retry_backoff_sleep.assert_not_awaited()

    async def test_setup_communication_with_pods_successful(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        task_id: str,
    ) -> None:
        """Setup communication successful."""
        expected_modeller_mailbox_id = "someMailboxId"
        expected_worker_mailboxes = {
            "somePodName": "somePodMailboxId",
            "someOtherPodName": "someOtherMailboxId",
        }
        grpc_stub.SetupTaskMailboxes.return_value = CommunicationDetails(
            mailboxId=expected_modeller_mailbox_id,
            podMailboxIds=expected_worker_mailboxes,
            taskId=task_id,
        )
        pod_identifier = "podOwner/somePodName"
        training_request = b"some training request"

        (
            modeller_mailbox_id,
            worker_mailboxes,
            received_task_id,
        ) = await message_service.setup_communication_with_pods(
            {pod_identifier: training_request}
        )

        assert modeller_mailbox_id == expected_modeller_mailbox_id
        assert worker_mailboxes == expected_worker_mailboxes
        assert received_task_id == task_id
        grpc_stub.SetupTaskMailboxes.assert_called_once_with(
            BitfountTasks(
                tasks=[
                    BitfountTask(
                        podIdentifier=pod_identifier, encryptedTask=training_request
                    )
                ]
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

    async def test_setup_communication_with_pods_successful_taskURL(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mocker: MockerFixture,
        task_id: str,
    ) -> None:
        """Setup communication successful with taskURL field."""
        expected_modeller_mailbox_id = "someMailboxId"
        expected_worker_mailboxes = {
            "somePodName": "somePodMailboxId",
            "someOtherPodName": "someOtherMailboxId",
        }
        grpc_stub.SetupTaskMailboxes.return_value = CommunicationDetails(
            mailboxId=expected_modeller_mailbox_id,
            podMailboxIds=expected_worker_mailboxes,
            taskId=task_id,
        )
        pod_identifier = "podOwner/somePodName"
        training_request = b"some training request"
        other_training_request = b"other training request"
        mock_task_upload = mocker.patch.object(
            _MessageService, "_maybe_upload_task_to_large_object_storage"
        )
        mock_task_upload.return_value = {pod_identifier: other_training_request}
        (
            modeller_mailbox_id,
            worker_mailboxes,
            received_task_id,
        ) = await message_service.setup_communication_with_pods(
            {pod_identifier: training_request}
        )

        assert modeller_mailbox_id == expected_modeller_mailbox_id
        assert worker_mailboxes == expected_worker_mailboxes
        assert received_task_id == task_id
        grpc_stub.SetupTaskMailboxes.assert_called_once_with(
            BitfountTasks(
                tasks=[
                    BitfountTask(
                        podIdentifier=pod_identifier, taskURL=other_training_request
                    )
                ]
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

    @pytest.mark.parametrize(
        argnames="status_code",
        argvalues=_GRPC_ERROR_CODES,
        ids=lambda sc: f"status_code={sc.name}",
    )
    async def test_setup_communication_with_pods_raises_error_on_failure(
        self,
        caplog: LogCaptureFixture,
        grpc_stub: Mock,
        message_service: _MessageService,
        remove_grpc_retry_backoff_sleep: AsyncMock,
        rpc_error_maker: RpcErrorMaker,
        status_code: StatusCode,
    ) -> None:
        """Setup communication receives RpcError."""
        if status_code in utils._RETRY_STATUS_CODES:
            retries = True
            expected_num_calls = utils._DEFAULT_MAX_RETRIES + 1
        else:
            retries = False
            expected_num_calls = 1
        expected_num_retries = expected_num_calls - 1

        grpc_stub.SetupTaskMailboxes.side_effect = rpc_error_maker(
            "SetupTaskMailboxes failed", status_code
        )
        pod_identifier = "podOwner/somePodName"
        training_request = b"some training request"

        with raises(
            BitfountMessageServiceError,
            match=re.escape(
                "Unable to setup communication with target pods."
                " See debug logs for more details."
            ),
        ), caplog.at_level(logging.DEBUG):
            await message_service.setup_communication_with_pods(
                {pod_identifier: training_request}
            )

        # Check calls
        assert (
            grpc_stub.SetupTaskMailboxes.call_args_list
            == [
                call(
                    BitfountTasks(
                        tasks=[
                            BitfountTask(
                                podIdentifier=pod_identifier,
                                encryptedTask=training_request,
                            )
                        ]
                    ),
                    metadata=[("token", ACCESS_TOKEN)],
                    timeout=utils._DEFAULT_TIMEOUT,
                )
            ]
            * expected_num_calls
        )

        # Check underlying exception was logged to DEBUG
        debug_logs = get_debug_logs(caplog, full_details=True)
        assert "RpcError: SetupTaskMailboxes failed" in debug_logs

        # Check retry messages in debug logs and sleep called if required
        if retries:
            assert remove_grpc_retry_backoff_sleep.await_count == expected_num_retries
            for i in range(1, expected_num_retries + 1):
                assert re.search(
                    rf"gRPC error occurred: .* SetupTaskMailboxes failed;"
                    rf" .* \(attempt {i}\)",
                    debug_logs,
                )
        else:
            remove_grpc_retry_backoff_sleep.assert_not_awaited()

    async def test_send_message_successful_already_packed(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        opt_task_id: Optional[str],
    ) -> None:
        """Test send packed body to Pod."""
        grpc_stub.SendBitfountMessage.return_value = SuccessResponse()
        packed_message = b"an already packed message"
        pod_identifier = "podOwner/somePodName"
        mailbox_id = "someMailboxId"
        reply_to_mailbox_id = "replyMailboxId"
        sender = "someSender"
        expected_timestamp = datetime.now().isoformat()

        response = await message_service.send_message(
            _BitfountMessage(
                message_type=_BitfountMessageType.TRAINING_UPDATE,
                body=packed_message,
                recipient=pod_identifier,
                recipient_mailbox_id=mailbox_id,
                sender=sender,
                sender_mailbox_id=reply_to_mailbox_id,
                timestamp=expected_timestamp,
                task_id=opt_task_id,
            ),
            already_packed=True,
        )

        assert response == SuccessResponse()
        grpc_stub.SendBitfountMessage.assert_called_once_with(
            GrpcBitfountMessage(
                messageType=_BitfountMessageType.TRAINING_UPDATE.value,
                body=packed_message,
                recipient=pod_identifier,
                recipientMailboxId=mailbox_id,
                sender=sender,
                senderMailboxId=reply_to_mailbox_id,
                timestamp=expected_timestamp,
                taskId=opt_task_id,
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

    @pytest.mark.parametrize(
        argnames="status_code",
        argvalues=_GRPC_ERROR_CODES,
        ids=lambda sc: f"status_code={sc.name}",
    )
    async def test_send_message_error_is_raised(
        self,
        caplog: LogCaptureFixture,
        grpc_stub: Mock,
        message_service: _MessageService,
        opt_task_id: Optional[str],
        remove_grpc_retry_backoff_sleep: AsyncMock,
        rpc_error_maker: RpcErrorMaker,
        status_code: StatusCode,
    ) -> None:
        """Test send to pod throws RpcError."""
        if status_code in utils._RETRY_STATUS_CODES:
            retries = True
            expected_num_calls = utils._DEFAULT_MAX_RETRIES + 1
        else:
            retries = False
            expected_num_calls = 1
        expected_num_retries = expected_num_calls - 1

        grpc_stub.SendBitfountMessage.side_effect = rpc_error_maker(
            "send message error", status_code
        )

        packed_message = b"an already packed message"
        pod_identifier = "podOwner/somePodName"
        mailbox_id = "someMailboxId"
        reply_to_mailbox_id = "replyMailboxId"
        sender = "someModeller"
        expected_timestamp = datetime.now().isoformat()

        with raises(
            BitfountMessageServiceError,
            match=re.escape(
                "Encountered problem when sending message."
                " See debug logs for more details."
            ),
        ), caplog.at_level(logging.DEBUG):
            await message_service.send_message(
                _BitfountMessage(
                    message_type=_BitfountMessageType.TRAINING_UPDATE,
                    body=packed_message,
                    recipient=pod_identifier,
                    recipient_mailbox_id=mailbox_id,
                    sender=sender,
                    sender_mailbox_id=reply_to_mailbox_id,
                    timestamp=expected_timestamp,
                    task_id=opt_task_id,
                ),
                already_packed=True,
            )

        # Check called the expected number of times
        assert (
            grpc_stub.SendBitfountMessage.call_args_list
            == [
                call(
                    GrpcBitfountMessage(
                        messageType=_BitfountMessageType.TRAINING_UPDATE.value,
                        body=packed_message,
                        sender=sender,
                        senderMailboxId=reply_to_mailbox_id,
                        recipient=pod_identifier,
                        recipientMailboxId=mailbox_id,
                        timestamp=expected_timestamp,
                        taskId=opt_task_id,
                    ),
                    metadata=[("token", ACCESS_TOKEN)],
                    timeout=utils._DEFAULT_TIMEOUT,
                )
            ]
            * expected_num_calls
        )

        # Check underlying exception was logged to DEBUG
        debug_logs = get_debug_logs(caplog, full_details=True)
        assert "RpcError: send message error" in debug_logs

        # Check retry messages in debug logs and sleep called if required
        if retries:
            assert remove_grpc_retry_backoff_sleep.await_count == expected_num_retries
            for i in range(1, expected_num_retries + 1):
                assert re.search(
                    rf"gRPC error occurred: .* send message error; .* \(attempt {i}\)",
                    debug_logs,
                )
        else:
            remove_grpc_retry_backoff_sleep.assert_not_awaited()

    async def test_send_message_successful_not_packed(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        opt_task_id: Optional[str],
    ) -> None:
        """Test send plain object to Modeller."""
        grpc_stub.SendBitfountMessage.return_value = SuccessResponse()
        unpacked_message = b"a message to be packed"
        expected_packed_message = msgpack.dumps(unpacked_message)
        modeller_name = "someModeller"
        pod_identifier = "podOwner/somePodName"
        mailbox_id = "someMailboxId"
        reply_to_mailbox_id = "replyMailboxId"
        expected_timestamp = datetime.now().isoformat()

        response = await message_service.send_message(
            _BitfountMessage(
                message_type=_BitfountMessageType.TRAINING_UPDATE,
                body=unpacked_message,
                recipient=modeller_name,
                recipient_mailbox_id=mailbox_id,
                sender=pod_identifier,
                sender_mailbox_id=reply_to_mailbox_id,
                timestamp=expected_timestamp,
                task_id=opt_task_id,
            ),
            already_packed=False,
        )

        assert response == SuccessResponse()
        grpc_stub.SendBitfountMessage.assert_called_once_with(
            GrpcBitfountMessage(
                messageType=_BitfountMessageType.TRAINING_UPDATE.value,
                body=expected_packed_message,
                sender=pod_identifier,
                senderMailboxId=reply_to_mailbox_id,
                recipient=modeller_name,
                recipientMailboxId=mailbox_id,
                timestamp=expected_timestamp,
                taskId=opt_task_id,
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

    async def test_get_message_successful(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mocker: MockerFixture,
        opt_task_id: Optional[str],
    ) -> None:
        """Test get message returns message."""
        expected_message_body = b"some message"
        expected_sent_by = "some modeller"
        expected_mailbox_id = "some mailbox"
        expected_reply_to_mailbox = "some reply to mailbox"
        expected_receipt_handle = "valid receipt handle"
        expected_pod_mailbox_ids = {"somePod": "someId"}
        expected_recipient = "someRecipient"
        expected_timestamp = datetime.now().isoformat()

        mock_sleep = mocker.patch("asyncio.sleep")

        grpc_stub.GetBitfountMessage.return_value = GrpcBitfountMessage(
            messageType=_BitfountMessageType.MODEL_PARAMETERS.value,
            body=expected_message_body,
            sender=expected_sent_by,
            senderMailboxId=expected_reply_to_mailbox,
            receiptHandle=expected_receipt_handle,
            recipient=expected_recipient,
            recipientMailboxId=expected_mailbox_id,
            podMailboxIds=expected_pod_mailbox_ids,
            timestamp=expected_timestamp,
            # Mimic when Grpc message doesn't have taskId
            **({"taskId": opt_task_id} if opt_task_id else {}),
        )
        grpc_stub.AcknowledgeMessage.return_value = SuccessResponse()

        message = await message_service._get_message(expected_mailbox_id)

        # Retrieved message is as expected
        assert message == _BitfountMessage(
            message_type=_BitfountMessageType.MODEL_PARAMETERS,
            body=expected_message_body,
            recipient=expected_recipient,
            recipient_mailbox_id=expected_mailbox_id,
            sender=expected_sent_by,
            sender_mailbox_id=expected_reply_to_mailbox,
            pod_mailbox_ids=expected_pod_mailbox_ids,
            timestamp=expected_timestamp,
            receipt_handle=expected_receipt_handle,
            task_id=opt_task_id,
        )

        # Message was requested correctly
        grpc_stub.GetBitfountMessage.assert_called_once_with(
            CommunicationDetails(mailboxId=expected_mailbox_id),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # Message was acknowledged correctly
        grpc_stub.AcknowledgeMessage.assert_called_once_with(
            Acknowledgement(
                mailboxId=expected_mailbox_id,
                receiptHandle=expected_receipt_handle,
                deleteMailbox=False,
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # We didn't sleep unnecessarily
        mock_sleep.assert_not_awaited()

    async def test_get_message_successful_and_is_unpacked(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mocker: MockerFixture,
        opt_task_id: Optional[str],
    ) -> None:
        """Test get message returns and unpacks message."""
        packed_message_body = msgpack.dumps(b"some message")
        expected_sent_by = "some modeller"
        expected_mailbox_id = "some mailbox"
        expected_reply_to_mailbox = "some reply to mailbox"
        expected_receipt_handle = "valid receipt handle"
        expected_recipient = "someRecipient"
        expected_timestamp = datetime.now().isoformat()

        mock_sleep = mocker.patch("asyncio.sleep")

        grpc_stub.GetBitfountMessage.return_value = GrpcBitfountMessage(
            messageType=_BitfountMessageType.MODEL_PARAMETERS.value,
            body=packed_message_body,
            sender=expected_sent_by,
            senderMailboxId=expected_reply_to_mailbox,
            receiptHandle=expected_receipt_handle,
            recipient=expected_recipient,
            recipientMailboxId=expected_mailbox_id,
            podMailboxIds={},
            timestamp=expected_timestamp,
            # Mimic when Grpc message doesn't have taskId
            **({"taskId": opt_task_id} if opt_task_id else {}),
        )
        grpc_stub.AcknowledgeMessage.return_value = SuccessResponse()

        message = await message_service._get_message(expected_mailbox_id)

        # Retrieved message is as expected
        assert message == _BitfountMessage(
            message_type=_BitfountMessageType.MODEL_PARAMETERS,
            body=packed_message_body,
            recipient=expected_recipient,
            recipient_mailbox_id=expected_mailbox_id,
            sender=expected_sent_by,
            sender_mailbox_id=expected_reply_to_mailbox,
            timestamp=expected_timestamp,
            receipt_handle=expected_receipt_handle,
            pod_mailbox_ids={},
            task_id=opt_task_id,
        )

        # Message was requested correctly
        grpc_stub.GetBitfountMessage.assert_called_once_with(
            CommunicationDetails(mailboxId=expected_mailbox_id),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # Message was acknowledged correctly
        grpc_stub.AcknowledgeMessage.assert_called_once_with(
            Acknowledgement(
                mailboxId=expected_mailbox_id,
                receiptHandle=expected_receipt_handle,
                deleteMailbox=False,
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # We didn't sleep unnecessarily
        mock_sleep.assert_not_awaited()

    async def test_get_message_successful_mailbox_delete_flag_set_on_task_complete(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mocker: MockerFixture,
        opt_task_id: Optional[str],
    ) -> None:
        """Test get message and delete mailbox.

        When a pod receives a TASK_COMPLETE message we acknowledge it,
        informing the message service that the task queue can now
        be deleted.
        """
        packed_message_body = msgpack.dumps(b"some message")
        expected_sent_by = "some modeller"
        expected_mailbox_id = "some mailbox"
        expected_reply_to_mailbox = "some reply to mailbox"
        expected_receipt_handle = "valid receipt handle"
        expected_recipient = "someRecipient"
        expected_timestamp = datetime.now().isoformat()

        mock_sleep = mocker.patch("asyncio.sleep")

        grpc_stub.GetBitfountMessage.return_value = GrpcBitfountMessage(
            messageType=_BitfountMessageType.TASK_COMPLETE.value,
            body=packed_message_body,
            sender=expected_sent_by,
            senderMailboxId=expected_reply_to_mailbox,
            receiptHandle=expected_receipt_handle,
            recipient=expected_recipient,
            recipientMailboxId=expected_mailbox_id,
            podMailboxIds={},
            timestamp=expected_timestamp,
            # Mimic when Grpc message doesn't have taskId
            **({"taskId": opt_task_id} if opt_task_id else {}),
        )
        grpc_stub.AcknowledgeMessage.return_value = SuccessResponse()

        message = await message_service._get_message(expected_mailbox_id)

        # Retrieved message is as expected
        assert message == _BitfountMessage(
            message_type=_BitfountMessageType.TASK_COMPLETE,
            body=packed_message_body,
            recipient=expected_recipient,
            recipient_mailbox_id=expected_mailbox_id,
            sender=expected_sent_by,
            sender_mailbox_id=expected_reply_to_mailbox,
            timestamp=expected_timestamp,
            receipt_handle=expected_receipt_handle,
            pod_mailbox_ids={},
            task_id=opt_task_id,
        )

        # Message was requested correctly
        grpc_stub.GetBitfountMessage.assert_called_once_with(
            CommunicationDetails(mailboxId=expected_mailbox_id),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # Message was acknowledged correctly
        grpc_stub.AcknowledgeMessage.assert_called_once_with(
            Acknowledgement(
                mailboxId=expected_mailbox_id,
                receiptHandle=expected_receipt_handle,
                deleteMailbox=True,
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # We didn't sleep unnecessarily
        mock_sleep.assert_not_awaited()

    async def test_get_message_downloads_from_large_object_storage(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mock_download_from_s3: AsyncMock,
        mocker: MockerFixture,
        opt_task_id: Optional[str],
        s3_download_url: _S3PresignedURL,
    ) -> None:
        """Test get message downloads from blob storage."""
        mock_async_sleep = mocker.patch("asyncio.sleep")

        # Construct GrpcBitfountMessage
        expected_message_body = s3_download_url
        packed_message_body = msgpack.dumps(expected_message_body)
        expected_sent_by = "some_modeller"
        expected_mailbox_id = "some_mailbox"
        expected_reply_to_mailbox = "some reply to mailbox"
        expected_receipt_handle = "valid receipt handle"
        expected_recipient = "some_recipient"
        expected_timestamp = datetime.now().isoformat()

        grpc_stub.GetBitfountMessage.return_value = GrpcBitfountMessage(
            messageType=_BitfountMessageType.TRAINING_UPDATE.value,
            body=packed_message_body,
            sender=expected_sent_by,
            senderMailboxId=expected_reply_to_mailbox,
            recipient=expected_recipient,
            recipientMailboxId=expected_mailbox_id,
            receiptHandle=expected_receipt_handle,
            timestamp=expected_timestamp,
            # Mimic when Grpc message doesn't have taskId
            **({"taskId": opt_task_id} if opt_task_id else {}),  # type: ignore[arg-type] # Reason: workaround to allow optional kwarg presence # noqa: B950
        )

        # Mock out message acknowledgement
        grpc_stub.AcknowledgeMessage.return_value = SuccessResponse()

        # Mock out downloading from S3
        # Messages are inherently packed when stored in S3, but we're mocking out
        # the return here so can avoid that.
        message_body_in_blob_storage = b"here is the actual message"
        mock_download_from_s3.return_value = message_body_in_blob_storage

        # Retrieve message
        message = await message_service._get_message(expected_mailbox_id)

        # Retrieved message is as expected
        assert (
            _BitfountMessage(
                message_type=_BitfountMessageType.TRAINING_UPDATE,
                body=message_body_in_blob_storage,
                recipient=expected_recipient,
                recipient_mailbox_id=expected_mailbox_id,
                sender=expected_sent_by,
                sender_mailbox_id=expected_reply_to_mailbox,
                timestamp=expected_timestamp,
                receipt_handle=expected_receipt_handle,
                pod_mailbox_ids={},
                task_id=opt_task_id,
            )
            == message
        )

        # Message was requested correctly
        grpc_stub.GetBitfountMessage.assert_called_once_with(
            CommunicationDetails(mailboxId=expected_mailbox_id),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # Message was acknowledged correctly
        grpc_stub.AcknowledgeMessage.assert_called_once_with(
            Acknowledgement(
                mailboxId=expected_mailbox_id,
                receiptHandle=expected_receipt_handle,
                deleteMailbox=False,
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # Download was called correctly
        mock_download_from_s3.assert_awaited_once_with(s3_download_url)

        # Check we didn't sleep unnecessarily
        mock_async_sleep.assert_not_awaited()

    async def test_get_message_none_found(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        remove_grpc_retry_backoff_sleep: AsyncMock,
        rpc_error_maker: RpcErrorMaker,
    ) -> None:
        """Test get message when there are none."""
        expected_mailbox_id = "some mailbox"

        error = rpc_error_maker("No message found", StatusCode.NOT_FOUND)

        grpc_stub.GetBitfountMessage.side_effect = error

        message = await message_service._get_message(expected_mailbox_id)

        # Retrieved message is as expected
        assert message is None

        # Message was requested correctly
        grpc_stub.GetBitfountMessage.assert_called_once_with(
            CommunicationDetails(mailboxId=expected_mailbox_id),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # Nothing to acknowledge
        grpc_stub.AcknowledgeMessage.assert_not_called()

        # We didn't sleep unnecessarily
        remove_grpc_retry_backoff_sleep.assert_not_awaited()

    @pytest.mark.parametrize(
        argnames="status_code",
        # We remove NOT_FOUND as this is tested separately
        argvalues=_GRPC_ERROR_CODES - {StatusCode.NOT_FOUND},
        ids=lambda sc: f"status_code={sc.name}",
    )
    async def test_get_message_throws_error_when_never_successful(
        self,
        caplog: LogCaptureFixture,
        grpc_stub: Mock,
        message_service: _MessageService,
        remove_grpc_retry_backoff_sleep: AsyncMock,
        rpc_error_maker: RpcErrorMaker,
        status_code: StatusCode,
    ) -> None:
        """Test get message eventually throws error if errors occur."""
        # Also need to remove NOT_FOUND from retryable status codes
        if status_code in utils._RETRY_STATUS_CODES:
            retries = True
            expected_num_calls = utils._DEFAULT_MAX_RETRIES + 1
        else:
            retries = False
            expected_num_calls = 1
        expected_num_retries = expected_num_calls - 1

        expected_mailbox_id = "some mailbox"

        grpc_stub.GetBitfountMessage.side_effect = rpc_error_maker(
            "an error", status_code
        )

        with raises(
            BitfountMessageServiceError,
            match=re.escape(
                "Issue retrieving message from mailbox. See debug log for details."
            ),
        ), caplog.at_level(logging.DEBUG):
            await message_service._get_message(expected_mailbox_id)

        # Message was requested correctly up to max attempts
        expected_get_call = call(
            CommunicationDetails(mailboxId=expected_mailbox_id),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )
        assert (
            grpc_stub.GetBitfountMessage.call_args_list
            == [expected_get_call] * expected_num_calls
        )

        # Nothing to acknowledge
        grpc_stub.AcknowledgeMessage.assert_not_called()

        # Check underlying exception was logged to DEBUG
        debug_logs = get_debug_logs(caplog, full_details=True)
        assert "RpcError: an error" in debug_logs

        # Check retry messages in debug logs and sleep called if required
        if retries:
            assert remove_grpc_retry_backoff_sleep.await_count == expected_num_retries
            for i in range(1, expected_num_retries + 1):
                assert re.search(
                    rf"gRPC error occurred: .* an error; .* \(attempt {i}\)",
                    debug_logs,
                )
        else:
            remove_grpc_retry_backoff_sleep.assert_not_awaited()

    async def test_get_message_throws_error_but_is_eventually_successful(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        opt_task_id: Optional[str],
        remove_grpc_retry_backoff_sleep: AsyncMock,
        rpc_error: RpcError,
    ) -> None:
        """Test get message doesn't throw error, retries until threshold is reached."""
        packed_message_body = msgpack.dumps(b"some message")
        expected_sent_by = "some modeller"
        expected_mailbox_id = "some mailbox"
        expected_reply_to_mailbox = "some reply to mailbox"
        expected_receipt_handle = "valid receipt handle"
        expected_recipient = "someRecipient"
        expected_timestamp = datetime.now().isoformat()

        grpc_stub.GetBitfountMessage.side_effect = [
            rpc_error,
            GrpcBitfountMessage(
                messageType=_BitfountMessageType.MODEL_PARAMETERS.value,
                body=packed_message_body,
                sender=expected_sent_by,
                senderMailboxId=expected_reply_to_mailbox,
                receiptHandle=expected_receipt_handle,
                recipient=expected_recipient,
                recipientMailboxId=expected_mailbox_id,
                podMailboxIds={},
                timestamp=expected_timestamp,
                # Mimic when Grpc message doesn't have taskId
                **({"taskId": opt_task_id} if opt_task_id else {}),
            ),
        ]
        grpc_stub.AcknowledgeMessage.return_value = SuccessResponse()

        message = await message_service._get_message(expected_mailbox_id)

        # Retrieved message is as expected
        assert message == _BitfountMessage(
            message_type=_BitfountMessageType.MODEL_PARAMETERS,
            body=packed_message_body,
            recipient=expected_recipient,
            recipient_mailbox_id=expected_mailbox_id,
            sender=expected_sent_by,
            sender_mailbox_id=expected_reply_to_mailbox,
            pod_mailbox_ids={},
            timestamp=expected_timestamp,
            receipt_handle=expected_receipt_handle,
            task_id=opt_task_id,
        )

        # Message was requested correctly
        expected_get_message_call = call(
            CommunicationDetails(mailboxId=expected_mailbox_id),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )
        grpc_stub.GetBitfountMessage.assert_has_calls([expected_get_message_call] * 2)
        assert grpc_stub.GetBitfountMessage.call_count == 2

        # Message was acknowledged correctly
        grpc_stub.AcknowledgeMessage.assert_called_once_with(
            Acknowledgement(
                mailboxId=expected_mailbox_id,
                receiptHandle=expected_receipt_handle,
                deleteMailbox=False,
            ),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )

        # We didn't sleep unnecessarily
        remove_grpc_retry_backoff_sleep.assert_awaited_once()

    async def test_get_message_is_called_again_if_acknowledgement_fails(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        opt_task_id: Optional[str],
        remove_grpc_retry_backoff_sleep: AsyncMock,
        rpc_error: RpcError,
    ) -> None:
        """Tests that the message is re-fetched if the acknowledgement fails.

        If acknowledgement fails then it may be due to an expired receipt handle,
        to avoid this we make sure that the message is retrieved again before retrying,
        as this will fetch a new receipt handle
        """
        packed_message_body = msgpack.dumps(b"some message")
        expected_sent_by = "some modeller"
        expected_mailbox_id = "some mailbox"
        expected_reply_to_mailbox = "some reply to mailbox"
        expected_receipt_handle = "invalid receipt handle"
        expected_second_receipt_handle = "valid receipt handle"
        expected_recipient = "someRecipient"
        expected_timestamp = datetime.now().isoformat()

        grpc_stub.GetBitfountMessage.side_effect = [
            GrpcBitfountMessage(
                messageType=_BitfountMessageType.MODEL_PARAMETERS.value,
                body=packed_message_body,
                sender=expected_sent_by,
                senderMailboxId=expected_reply_to_mailbox,
                receiptHandle=expected_receipt_handle,
                recipient=expected_recipient,
                recipientMailboxId=expected_mailbox_id,
                podMailboxIds={},
                timestamp=expected_timestamp,
                # Mimic when Grpc message doesn't have taskId
                **({"taskId": opt_task_id} if opt_task_id else {}),
            ),
            GrpcBitfountMessage(
                messageType=_BitfountMessageType.MODEL_PARAMETERS.value,
                body=packed_message_body,
                sender=expected_sent_by,
                senderMailboxId=expected_reply_to_mailbox,
                receiptHandle=expected_second_receipt_handle,
                recipient=expected_recipient,
                recipientMailboxId=expected_mailbox_id,
                podMailboxIds={},
                timestamp=expected_timestamp,
                # Mimic when Grpc message doesn't have taskId
                **({"taskId": opt_task_id} if opt_task_id else {}),
            ),
        ]

        grpc_stub.AcknowledgeMessage.side_effect = [rpc_error, SuccessResponse()]

        message = await message_service._get_message(expected_mailbox_id)

        # Retrieved message is as expected
        assert message == _BitfountMessage(
            message_type=_BitfountMessageType.MODEL_PARAMETERS,
            body=packed_message_body,
            recipient=expected_recipient,
            recipient_mailbox_id=expected_mailbox_id,
            sender=expected_sent_by,
            sender_mailbox_id=expected_reply_to_mailbox,
            pod_mailbox_ids={},
            timestamp=expected_timestamp,
            receipt_handle=expected_second_receipt_handle,
            task_id=opt_task_id,
        )

        # Message was requested correctly
        expected_get_message_call = call(
            CommunicationDetails(mailboxId=expected_mailbox_id),
            metadata=[("token", ACCESS_TOKEN)],
            timeout=utils._DEFAULT_TIMEOUT,
        )
        grpc_stub.GetBitfountMessage.assert_has_calls([expected_get_message_call] * 2)
        assert grpc_stub.GetBitfountMessage.call_count == 2

        # Message was acknowledged correctly
        grpc_stub.AcknowledgeMessage.assert_has_calls(
            [
                call(
                    Acknowledgement(
                        mailboxId=expected_mailbox_id,
                        receiptHandle=expected_receipt_handle,
                        deleteMailbox=False,
                    ),
                    metadata=[("token", ACCESS_TOKEN)],
                    timeout=utils._DEFAULT_TIMEOUT,
                ),
                call(
                    Acknowledgement(
                        mailboxId=expected_mailbox_id,
                        receiptHandle=expected_second_receipt_handle,
                        deleteMailbox=False,
                    ),
                    metadata=[("token", ACCESS_TOKEN)],
                    timeout=utils._DEFAULT_TIMEOUT,
                ),
            ]
        )
        assert grpc_stub.AcknowledgeMessage.call_count == 2

        # Sleep was called between errors
        remove_grpc_retry_backoff_sleep.assert_awaited_once()

    async def test_poll_for_message_eventually_retrieves_message(
        self,
        message_service: _MessageService,
        mocker: MockerFixture,
        opt_task_id: Optional[str],
    ) -> None:
        """Test poll for message receives a message after a few attempts."""
        expected_message = _BitfountMessage(
            message_type=_BitfountMessageType.UNDEFINED,
            body=b"hello world",
            recipient="to_you",
            recipient_mailbox_id="your_mailbox_id",
            sender="from_me",
            sender_mailbox_id="my_mailbox_id",
            task_id=opt_task_id,
        )
        expected_mailbox = "someMailboxID"
        mocked__get_message = mocker.patch.object(
            message_service,
            "_get_message",
            AsyncMock(
                side_effect=[
                    None,
                    None,
                    expected_message,
                ]
            ),
        )

        # There is no anext() built-in so have to do it this way
        message = await message_service.poll_for_messages(
            expected_mailbox, stop_event=Mock(**{"is_set.return_value": False})
        ).__anext__()

        assert message == expected_message
        mocked__get_message.assert_has_awaits([call(expected_mailbox)] * 3)
        assert mocked__get_message.await_count == 3

    async def test_poll_for_message_throws_error(
        self,
        message_service: _MessageService,
        mocker: MockerFixture,
        rpc_error: RpcError,
    ) -> None:
        """Test poll for message throws RpcError."""
        expected_mailbox = "someMailboxID"
        mocked__get_message = mocker.patch.object(
            message_service, "_get_message", AsyncMock(side_effect=[None, rpc_error])
        )

        with raises(RpcError):
            # There is no anext() built-in so have to do it this way
            await message_service.poll_for_messages(
                expected_mailbox, stop_event=Mock(**{"is_set.return_value": False})
            ).__anext__()

        mocked__get_message.assert_has_awaits([call(expected_mailbox)] * 2)
        assert mocked__get_message.await_count == 2

    async def test_maybe_upload_to_large_object_storage_small_message(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mock_upload_to_s3: AsyncMock,
        opt_task_id: Optional[str],
        s3_download_url: _S3PresignedURL,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Test small messages are sent to message service."""
        # Set output as expected fields
        grpc_stub.GetLargeObjectStorage.return_value = BlobStorageData(
            uploadUrl=s3_upload_url,
            downloadUrl=s3_download_url,
            uploadFields=s3_upload_fields,
        )

        # Construct message
        sender = "someSender"
        sender_mailbox_id = "someSenderMailboxId"
        expected_body = msgpack.dumps({"some": "body"})
        bitfount_message = _BitfountMessage(
            body=expected_body,  #
            message_type=_BitfountMessageType.UNDEFINED,
            recipient="someRecipient",
            recipient_mailbox_id="someMailboxId",
            sender=sender,
            sender_mailbox_id=sender_mailbox_id,
            task_id=opt_task_id,
        )

        body = await message_service._maybe_upload_to_large_object_storage(
            bitfount_message
        )

        # Check the upload function wasn't awaited
        mock_upload_to_s3.assert_not_awaited()
        # Check that message body is unchanged
        assert body == expected_body

    async def test_maybe_upload_task_to_large_object_storage_small_message(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mock_upload_to_s3: AsyncMock,
        opt_task_id: Optional[str],
        s3_download_url: _S3PresignedURL,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Test small messages are sent to message service."""
        # Set output as expected fields
        grpc_stub.GetLargeObjectStorage.return_value = BlobStorageData(
            uploadUrl=s3_upload_url,
            downloadUrl=s3_download_url,
            uploadFields=s3_upload_fields,
        )

        tasks_per_pod = {"pod1": b"some task"}

        new_task = await message_service._maybe_upload_task_to_large_object_storage(
            tasks_per_pod
        )

        # Check the upload function wasn't awaited
        mock_upload_to_s3.assert_not_awaited()
        # check that the tasks are not modified
        assert new_task == tasks_per_pod

    async def test_maybe_upload_task_to_large_object_storage_large_message(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mock_upload_to_s3: AsyncMock,
        opt_task_id: Optional[str],
        s3_download_url: _S3PresignedURL,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Test large task object gets uploaded."""
        # Set output as expected fields
        grpc_stub.GetLargeObjectStorage.return_value = BlobStorageData(
            uploadUrl=s3_upload_url,
            downloadUrl=s3_download_url,
            uploadFields=s3_upload_fields,
        )

        # Construct tasks
        expected_body = b"a" * (_SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES + 1)
        tasks_per_pod = {"someSender/somePod": expected_body}
        message_service._session._client_id = "id"
        new_tasks = await message_service._maybe_upload_task_to_large_object_storage(
            tasks_per_pod
        )

        # Assert upload code is called
        mock_upload_to_s3.assert_awaited_once_with(
            upload_url=s3_upload_url,
            presigned_fields=s3_upload_fields,
            data=expected_body,
        )
        # Assert download URL is returned (packed) to user
        assert new_tasks["someSender/somePod"] == msgpack.dumps(s3_download_url)

    async def test_maybe_upload_task_to_large_object_storage_fails_too_large_message(
        self, message_service: _MessageService, mocker: MockerFixture
    ) -> None:
        """Tests that exception raised if message body too big for storage."""
        # Patch out schema size calculation so we can set it
        too_large_size = _MAX_STORAGE_SIZE_BYTES + 1
        mock_data_sizer = mocker.patch(
            "bitfount.federated.transport.message_service._get_packed_data_object_size"
        )
        mock_data_sizer = mocker.patch(
            "bitfount.federated.transport.message_service._get_packed_data_object_size"
        )
        mock_data_sizer.return_value = too_large_size

        tasks_per_pod = {"pod1": Mock()}
        tasks_per_pod["pod1"].__len__ = lambda _: too_large_size

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Message body is too large to upload: "
                f"expected max {_MAX_STORAGE_SIZE_MEGABYTES} megabytes, "
                f"got {_get_mb_from_bytes(too_large_size).fractional} megabytes."
            ),
        ):
            await message_service._maybe_upload_task_to_large_object_storage(
                tasks_per_pod  # type: ignore[arg-type] # Reason: Mocking for testing purposes # noqa: B950
            )

    async def test_maybe_upload_to_large_object_storage_large_message(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mock_upload_to_s3: AsyncMock,
        opt_task_id: Optional[str],
        s3_download_url: _S3PresignedURL,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Test large object gets uploaded."""
        # Set output as expected fields
        grpc_stub.GetLargeObjectStorage.return_value = BlobStorageData(
            uploadUrl=s3_upload_url,
            downloadUrl=s3_download_url,
            uploadFields=s3_upload_fields,
        )

        # Construct message
        sender = "someSender/somePod"
        sender_mailbox_id = "someSenderMailboxId"
        expected_body = b"a" * (_SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES + 1)
        bitfount_message = _BitfountMessage(
            body=expected_body,
            message_type=_BitfountMessageType.UNDEFINED,
            recipient="someRecipient",
            recipient_mailbox_id="someMailboxId",
            sender=sender,
            sender_mailbox_id=sender_mailbox_id,
            task_id=opt_task_id,
        )

        body = await message_service._maybe_upload_to_large_object_storage(
            bitfount_message
        )

        # Assert upload code is called
        mock_upload_to_s3.assert_awaited_once_with(
            upload_url=s3_upload_url,
            presigned_fields=s3_upload_fields,
            data=expected_body,
        )
        # Assert download URL is returned (packed) to user
        assert body == msgpack.dumps(s3_download_url)

    async def test_maybe_upload_to_large_object_storage_fails_too_large_message(
        self, message_service: _MessageService, mocker: MockerFixture
    ) -> None:
        """Tests that exception raised if message body too big for storage."""
        # Patch out schema size calculation so we can set it
        too_large_size = _MAX_STORAGE_SIZE_BYTES + 1
        mock_data_sizer = mocker.patch(
            "bitfount.federated.transport.message_service._get_packed_data_object_size"
        )
        mock_data_sizer.return_value = too_large_size

        # Create mock message, mock length is long enough to consider uploading
        mock_message = create_dataclass_mock(_BitfountMessage)
        mock_message.body.__len__ = lambda _: too_large_size

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Message body is too large to upload: "
                f"expected max {_MAX_STORAGE_SIZE_MEGABYTES} megabytes, "
                f"got {_get_mb_from_bytes(too_large_size).fractional} megabytes."
            ),
        ):
            await message_service._maybe_upload_to_large_object_storage(mock_message)

    async def test_maybe_upload_to_large_object_storage_bad_upload(
        self,
        grpc_stub: Mock,
        message_service: _MessageService,
        mock_upload_to_s3: AsyncMock,
        opt_task_id: Optional[str],
        s3_download_url: _S3PresignedURL,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Test failed upload throws exception."""
        # Set output as expected fields
        grpc_stub.GetLargeObjectStorage.return_value = BlobStorageData(
            uploadUrl=s3_upload_url,
            downloadUrl=s3_download_url,
            uploadFields=s3_upload_fields,
        )

        # Construct message
        sender = "someSender"
        sender_mailbox_id = "someSenderMailboxId"
        expected_body = b"a" * (_SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES + 1)
        bitfount_message = _BitfountMessage(
            body=expected_body,
            message_type=_BitfountMessageType.UNDEFINED,
            recipient="someRecipient",
            recipient_mailbox_id="someMailboxId",
            sender=sender,
            sender_mailbox_id=sender_mailbox_id,
            task_id=opt_task_id,
        )

        # Set upload mock to throw exception
        mock_upload_to_s3.side_effect = HTTPError("TEST ERROR")

        with pytest.raises(
            RequestException,
            match=re.escape(
                "Failed to upload message to large message storage. Cause: TEST ERROR."
            ),
        ):
            await message_service._maybe_upload_to_large_object_storage(
                bitfount_message
            )

        # Assert upload function called correctly even if errored
        mock_upload_to_s3.assert_awaited_once_with(
            upload_url=s3_upload_url,
            presigned_fields=s3_upload_fields,
            data=expected_body,
        )

    @pytest.mark.parametrize(
        argnames="status_code",
        argvalues=_GRPC_ERROR_CODES,
        ids=lambda sc: f"status_code={sc.name}",
    )
    async def test_maybe_upload_to_large_object_storage_rpc_error(
        self,
        caplog: LogCaptureFixture,
        grpc_stub: Mock,
        message_service: _MessageService,
        opt_task_id: Optional[str],
        remove_grpc_retry_backoff_sleep: AsyncMock,
        rpc_error_maker: RpcErrorMaker,
        status_code: StatusCode,
    ) -> None:
        """Test failed upload throws exception."""
        if status_code in utils._RETRY_STATUS_CODES:
            retries = True
            expected_num_calls = utils._DEFAULT_MAX_RETRIES + 1
        else:
            retries = False
            expected_num_calls = 1
        expected_num_retries = expected_num_calls - 1

        # Make GRPC error out
        grpc_stub.GetLargeObjectStorage.side_effect = rpc_error_maker(
            "Failed to create storage", status_code
        )

        # Construct message that should be uploaded
        sender = "someSender"
        sender_mailbox_id = "someSenderMailboxId"
        expected_body = b"a" * (_SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES + 1)
        bitfount_message = _BitfountMessage(
            body=expected_body,
            message_type=_BitfountMessageType.UNDEFINED,
            recipient="someRecipient",
            recipient_mailbox_id="someMailboxId",
            sender=sender,
            sender_mailbox_id=sender_mailbox_id,
            task_id=opt_task_id,
        )

        with pytest.raises(
            BitfountMessageServiceError,
            match=re.escape(
                "Unable to acquire large object storage."
                " See debug logs for more details."
            ),
        ), caplog.at_level(logging.DEBUG):
            await message_service._maybe_upload_to_large_object_storage(
                bitfount_message
            )

        # Check underlying exception was logged to DEBUG
        debug_logs = get_debug_logs(caplog, full_details=True)
        assert "RpcError: Failed to create storage" in debug_logs

        # Check retry messages in debug logs and sleep called if required
        if retries:
            assert remove_grpc_retry_backoff_sleep.await_count == expected_num_retries
            for i in range(1, expected_num_retries + 1):
                assert re.search(
                    rf"gRPC error occurred: .* Failed to create storage;"
                    rf" .* \(attempt {i}\)",
                    debug_logs,
                )
        else:
            remove_grpc_retry_backoff_sleep.assert_not_awaited()

    async def test_upload_large_object_bad_response(
        self,
        mock_upload_to_s3: AsyncMock,
        s3_upload_fields: _S3PresignedPOSTFields,
        s3_upload_url: _S3PresignedPOSTURL,
    ) -> None:
        """Uploading fails due to non 200 response code."""
        # Set upload mock to throw error (non-200 response code)
        mock_upload_to_s3.side_effect = HTTPError("NON-200/201 TEST")
        object_to_upload = b"here is my byte string"

        with pytest.raises(
            RequestException,
            match=re.escape(
                "Failed to upload message to large message storage. "
                "Cause: NON-200/201 TEST."
            ),
        ):
            await _LargeObjectRequestHandler.upload_large_object(
                s3_upload_url, s3_upload_fields, object_to_upload
            )

        # Assert upload function awaited correctly even if errored
        mock_upload_to_s3.assert_awaited_once_with(
            upload_url=s3_upload_url,
            presigned_fields=s3_upload_fields,
            data=object_to_upload,
        )

    async def test_download_large_object_bad_response(
        self, mock_download_from_s3: AsyncMock, s3_download_url: _S3PresignedURL
    ) -> None:
        """Downloading fails due to non 200 response code."""
        # Set mock to return HTTP Error (non-200 response code)
        mock_download_from_s3.side_effect = HTTPError("NON-200 TEST ERROR")

        with pytest.raises(
            RequestException,
            match=re.escape(
                "Failed to retrieve message from large message storage. "
                "Cause: NON-200 TEST ERROR."
            ),
        ):
            await _LargeObjectRequestHandler.get_large_object_from_url(s3_download_url)

        # Assert upload function called correctly even if errored
        mock_download_from_s3.assert_awaited_once_with(download_url=s3_download_url)

    def test_save_object_to_local_storage(
        self,
        message_service: _MessageService,
        opt_task_id: Optional[str],
    ) -> None:
        """Tests saving object to local storage."""
        sender = "someSender"
        sender_mailbox_id = "someSenderMailboxId"
        expected_body = b"a" * (_SMALL_MESSAGE_UPPER_LIMIT_SIZE_BYTES + 1)

        bitfount_message = _BitfountMessage(
            body=expected_body,
            message_type=_BitfountMessageType.UNDEFINED,
            recipient="someRecipient",
            recipient_mailbox_id="someMailboxId",
            sender=sender,
            sender_mailbox_id=sender_mailbox_id,
            task_id=opt_task_id,
        )

        filename = message_service._save_object_to_local_storage(bitfount_message)
        assert type(filename) is bytes
        filename_str: str = msgpack.loads(filename)
        assert Path(filename_str).exists()
        with open(filename_str, "rb") as f:
            assert f.read() == expected_body

    async def test_read_local_bitfount_message_from_rpc(
        self, opt_task_id: Optional[str]
    ) -> None:
        """Test reading local object from storage."""
        sender = "someSender"
        mailbox_id = "someSenderMailboxId"
        handle, local_tempfile = tempfile.mkstemp()
        packed_message = msgpack.dumps(local_tempfile)
        expected_message_body = msgpack.dumps(["some message"])
        with os.fdopen(handle, "wb") as f:
            f.write(expected_message_body)
        pod_identifier = "podOwner/somePodName"
        reply_to_mailbox_id = "replyMailboxId"
        expected_timestamp = datetime.now().isoformat()

        grpc_message = GrpcBitfountMessage(
            messageType=_BitfountMessageType.TRAINING_UPDATE.value,
            body=packed_message,
            recipient=pod_identifier,
            recipientMailboxId=mailbox_id,
            sender=sender,
            senderMailboxId=reply_to_mailbox_id,
            timestamp=expected_timestamp,
            # Mimic when Grpc message doesn't have taskId
            **({"taskId": opt_task_id} if opt_task_id else {}),  # type: ignore[arg-type] # Reason: workaround to allow optional kwarg presence # noqa: B950
        )
        bitfount_message = await _BitfountMessage.from_rpc(grpc_message)
        assert type(bitfount_message) == _BitfountMessage
        assert bitfount_message.body == expected_message_body


@unit_test
class TestMessageEncryption:
    """Tests AES message encryption."""

    @fixture
    def original_message(self) -> bytes:
        """Message to encrypt."""
        return b"this is my original message"

    @fixture
    def original_dataframe(self) -> pd.DataFrame:
        """Create DataFrame message to encrypt."""
        data = create_dataset(classification=True)
        return data

    @fixture
    def original_numpyarray(self) -> np.ndarray:
        """Create Numpt Array to encrypt."""
        data = create_dataset(classification=True)
        return cast(np.ndarray, data["A"].to_numpy())

    def test_encryption_is_polymorphic(self, original_message: bytes) -> None:
        """Testing that we can decrypt an encrypted message."""
        encryption_key = _AESEncryption.generate_key()

        assert (
            _MessageEncryption.decrypt_incoming_message(
                _MessageEncryption.encrypt_outgoing_message(
                    original_message, encryption_key
                ),
                encryption_key,
            )
            == original_message
        )

    def test_encode_dataframe(self, original_dataframe: pd.DataFrame) -> None:
        """Testing that we can encode a dataframe."""
        # Encode the DataFrame
        encoded_dataframe = msgpack.dumps(original_dataframe, default=msgpackext_encode)
        assert len(encoded_dataframe) > 0

    def test_encode_numpyarray(self, original_numpyarray: np.ndarray) -> None:
        """Testing that we can encode a Numpy Array."""
        # Encode the Numpy Array
        encoded_dataframe = msgpack.dumps(
            original_numpyarray, default=msgpackext_encode
        )
        assert len(encoded_dataframe) > 0

    def test_encryption_is_polymorphic_dataframe(
        self, original_dataframe: pd.DataFrame
    ) -> None:
        """Testing that we can decrypt an encrypted DataFrame."""
        encryption_key = _AESEncryption.generate_key()

        # Encode and encrypt the DataFrame
        encoded_dataframe = msgpack.dumps(original_dataframe, default=msgpackext_encode)
        encrypted_message = _MessageEncryption.encrypt_outgoing_message(
            encoded_dataframe, encryption_key
        )
        # Decrypt and decode the DataFrame
        decrypted_message = _MessageEncryption.decrypt_incoming_message(
            encrypted_message, encryption_key
        )
        decoded_dataframe = msgpack.loads(decrypted_message, ext_hook=msgpackext_decode)

        pd.testing.assert_frame_equal(original_dataframe, decoded_dataframe)

    def test_encryption_is_polymorphic_numpyarray(
        self, original_numpyarray: np.ndarray
    ) -> None:
        """Testing that we can decrypt an encrypted Numpy Array."""
        encryption_key = _AESEncryption.generate_key()

        # Encode and encrypt the Numpy Array
        encoded_array = msgpack.dumps(original_numpyarray, default=msgpackext_encode)
        encrypted_message = _MessageEncryption.encrypt_outgoing_message(
            encoded_array, encryption_key
        )
        # Decrypt and decode the Numpy Array
        decrypted_message = _MessageEncryption.decrypt_incoming_message(
            encrypted_message, encryption_key
        )
        decoded_numpyarray = msgpack.loads(
            decrypted_message, ext_hook=msgpackext_decode
        )

        assert np.array_equal(original_numpyarray, decoded_numpyarray)

    def test_encrypted_message_is_different(self, original_message: bytes) -> None:
        """Sanity checking we haven't just returned the message."""
        encryption_key = _AESEncryption.generate_key()

        encrypted_message = _MessageEncryption.encrypt_outgoing_message(
            original_message, encryption_key
        )

        assert encrypted_message != original_message

    def test_provided_encryption_key_is_used(self, original_message: bytes) -> None:
        """Sanity checking the key actually serves a purpose."""
        encrypted_message = _MessageEncryption.encrypt_outgoing_message(
            original_message, _AESEncryption.generate_key()
        )

        wrong_encryption_key = _AESEncryption.generate_key()

        with pytest.raises(
            DecryptError, match=re.escape("Unable to decrypt ciphertext")
        ):
            _MessageEncryption.decrypt_incoming_message(
                encrypted_message, wrong_encryption_key
            )


@unit_test
class TestBitfountMessage:
    """Tests for BitfountMessage."""

    def test_timestamps_unique(self) -> None:
        """Tests the timestamps for two BitfountMessages are different.

        This will help avoid a regression bug where we had the same default
        timestamp for every instance of the class due the nature of default arg
        values in dataclasses.
        """
        bm1 = _BitfountMessage(
            message_type=Mock(),
            body=b"",
            recipient="recipient",
            recipient_mailbox_id="recipient_mailbox_id",
            sender="sender",
            sender_mailbox_id="sender_mailbox_id",
            task_id="task_id",
        )
        time.sleep(1)  # Windows gives error on this test without the sleep
        bm2 = _BitfountMessage(
            message_type=Mock(),
            body=b"",
            recipient="recipient",
            recipient_mailbox_id="recipient_mailbox_id",
            sender="sender",
            sender_mailbox_id="sender_mailbox_id",
            task_id="task_id",
        )
        assert bm1.timestamp != bm2.timestamp

    def test_decrypt_rsa(self) -> None:
        """Test that decrypt_rsa() method works."""
        # Create RSA encrypted message
        msg_contents = "Hello, world!"
        private_key, public_key = _RSAEncryption.generate_key_pair()
        encypted_msg_contents = _RSAEncryption.encrypt(
            msgpack.dumps(msg_contents), public_key
        )

        # Create message with RSA encrypted body
        bitfount_message = _BitfountMessage(
            message_type=Mock(),
            body=encypted_msg_contents,
            recipient=Mock(),
            recipient_mailbox_id=Mock(),
            sender=Mock(),
            sender_mailbox_id=Mock(),
            task_id=Mock(),
        )
        decrypted_bitfount_message = bitfount_message.decrypt_rsa(private_key)

        # Check decrypted body is same as original contents
        assert decrypted_bitfount_message.body == msg_contents
        # Check other attributes are unchanged
        assert decrypted_bitfount_message.message_type == bitfount_message.message_type
        assert decrypted_bitfount_message.recipient == bitfount_message.recipient
        assert (
            decrypted_bitfount_message.recipient_mailbox_id
            == bitfount_message.recipient_mailbox_id
        )
        assert decrypted_bitfount_message.sender == bitfount_message.sender
        assert (
            decrypted_bitfount_message.sender_mailbox_id
            == bitfount_message.sender_mailbox_id
        )
        assert (
            decrypted_bitfount_message.pod_mailbox_ids
            == bitfount_message.pod_mailbox_ids
        )
        assert decrypted_bitfount_message.timestamp == bitfount_message.timestamp
        assert decrypted_bitfount_message.task_id == bitfount_message.task_id

    def test_default_task_id(self) -> None:
        """Tests that if a task_id is not specified, it is None."""
        bm = _BitfountMessage(
            message_type=Mock(),
            body=b"",
            recipient="recipient",
            recipient_mailbox_id="recipient_mailbox_id",
            sender="sender",
            sender_mailbox_id="sender_mailbox_id",
        )
        assert bm.task_id is None

    async def test_from_rpc_replaces_None_task_id(self) -> None:
        """Tests that if GRPC taskId is None, task_id is None."""
        grpc_message = GrpcBitfountMessage(
            messageType=_BitfountMessageType.JOB_ACCEPT.value,
            body=b"Hello world!",
            recipient="recipient",
            recipientMailboxId="recipientMailboxId",
            sender="sender",
            senderMailboxId="senderMailboxId",
            timestamp=time.asctime(),
            taskId=None,
        )
        bm = await _BitfountMessage.from_rpc(grpc_message)
        assert bm.task_id is None

    async def test_from_rpc_replaces_empty_task_id(self) -> None:
        """Tests that if GRPC taskId is not specified, task_id is None."""
        grpc_message = GrpcBitfountMessage(
            messageType=_BitfountMessageType.JOB_ACCEPT.value,
            body=b"Hello world!",
            recipient="recipient",
            recipientMailboxId="recipientMailboxId",
            sender="sender",
            senderMailboxId="senderMailboxId",
            timestamp=time.asctime(),
        )
        bm = await _BitfountMessage.from_rpc(grpc_message)
        assert bm.task_id is None


@unit_test
class TestDecryptedBitfountMessage:
    """Tests for BitfountMessage."""

    def test_timestamps_unique(self) -> None:
        """Tests the timestamps for two DecryptedBitfountMessages are different.

        This will help avoid a regression bug where we had the same default
        timestamp for every instance of the class due the nature of default arg
        values in dataclasses.
        """
        dbm1 = _DecryptedBitfountMessage(
            message_type=Mock(),
            body=Mock(),
            recipient="recipient",
            recipient_mailbox_id="recipient_mailbox_id",
            sender="sender",
            sender_mailbox_id="sender_mailbox_id",
            task_id="task_id",
        )
        time.sleep(1)  # Windows gives error on this test without the sleep
        dbm2 = _DecryptedBitfountMessage(
            message_type=Mock(),
            body=Mock(),
            recipient="recipient",
            recipient_mailbox_id="recipient_mailbox_id",
            sender="sender",
            sender_mailbox_id="sender_mailbox_id",
            task_id="task_id",
        )
        assert dbm1.timestamp != dbm2.timestamp


@unit_test
def test_msgpackext_encode_returns_obj() -> None:
    """Tests msgpack_encode returns same obj if not np or pd."""
    obj = Mock()
    assert msgpackext_encode(obj) == obj


@unit_test
def test_msgpackext_decode_returns_obj() -> None:
    """Tests msgpack_decode returns same obj if not np or pd."""
    obj = Mock()
    assert msgpackext_decode(code=0, obj=obj) == obj



"""Tests for the model evaluation algorithm."""
from typing import TYPE_CHECKING
from unittest.mock import Mock, create_autospec

import pytest
from pytest_mock import MockerFixture

from bitfount import BitfountSchema
from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
from bitfount.federated.algorithms.base import (
    _BaseAlgorithm,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.algorithms.model_algorithms import evaluate
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithm,
    _BaseModellerModelAlgorithm,
    _BaseWorkerModelAlgorithm,
)
from bitfount.federated.algorithms.model_algorithms.evaluate import (
    ModelEvaluation,
    _ModellerSide,
    _WorkerSide,
)
from bitfount.federated.utils import _ALGORITHMS
from bitfount.hub import BitfountHub
from bitfount.schemas.utils import bf_dump, bf_load
from tests.bitfount.backends.pytorch.models.test_models import assert_vars_equal
from tests.utils.helper import backend_test, create_datastructure, unit_test


class TestModelEvaluation:
    """Test Evaluate algorithm."""

    @unit_test
    def test_modeller_types(self, model: Mock) -> None:
        """Test modeller method."""
        algorithm_factory = ModelEvaluation(model=model)
        algorithm = algorithm_factory.modeller()
        for type_ in [
            _BaseAlgorithm,
            _BaseModellerAlgorithm,
            _BaseModelAlgorithm,
            _BaseModellerModelAlgorithm,
        ]:
            assert isinstance(algorithm, type_)

    @unit_test
    def test_worker_types(self, model: Mock) -> None:
        """Test worker method."""
        algorithm_factory = ModelEvaluation(model=model)
        algorithm = algorithm_factory.worker(
            hub=create_autospec(BitfountHub, instance=True)
        )
        for type_ in [
            _BaseAlgorithm,
            _BaseWorkerAlgorithm,
            _BaseModelAlgorithm,
            _BaseWorkerModelAlgorithm,
        ]:
            assert isinstance(algorithm, type_)

    @unit_test
    def test_attribute_error_raised_run(self) -> None:
        """Test non abstract class without run method raises AttributeError."""
        err_msg = "does not have a `run` method"
        for type_ in [
            _BaseWorkerAlgorithm,
            _BaseWorkerModelAlgorithm,
        ]:
            with pytest.raises(AttributeError) as e_info:

                class NoRunMethod(type_):  # type: ignore[valid-type, misc] # reason: test child methods need run method # noqa: B950
                    def initialise(self) -> None:
                        pass

            assert err_msg in str(e_info)

    @unit_test
    def test_attribute_error_raised_initialise(self) -> None:
        """Test non abstract class without run method raises AttributeError."""
        err_msg = "does not have a `initialise` method"

        # All model algorithms have the initialise method on the
        # base class so we only test the `_BaseWorkerAlgorithm`
        with pytest.raises(AttributeError) as e_info:

            class NoInitialiseMethod(_BaseWorkerAlgorithm):
                def run(self) -> None:
                    pass

        assert err_msg in str(e_info)

    @unit_test
    def test_worker_run(self, mocker: MockerFixture, model: Mock) -> None:
        """Tests that worker run does metric calculation."""
        worker = evaluate._WorkerSide(model=model)
        mock_metrics = Mock()
        mocker.patch(
            "bitfount.metrics.MetricCollection.create_from_model", mock_metrics
        )
        worker.run()
        model.evaluate.assert_called_once()
        mock_metrics.assert_called_once()

    @unit_test
    def test_modeller_run(self, model: Mock) -> None:
        """Tests that modeller run returns results."""
        modeller = evaluate._ModellerSide(model=model)
        results = {"pod1": {"AUC": 0.5}}
        assert results == modeller.run(results=results)


@backend_test
@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for column average algorithm."""

    def test_serialization(self) -> None:
        """Test dump and load for ModelEvaluation."""
        model = PyTorchTabularClassifier(
            datastructure=create_datastructure(),
            schema=BitfountSchema(),
            epochs=2,
        )
        algorithm_factory = ModelEvaluation(model=model)
        dumped = bf_dump(algorithm_factory)
        loaded = bf_load(dumped, _ALGORITHMS)

        assert algorithm_factory.class_name == loaded.class_name
        assert_vars_equal(vars(algorithm_factory.model), vars(loaded.model))


# Static tests for algorithm-protocol compatibility
if TYPE_CHECKING:
    from typing import cast

    from bitfount.federated.protocols.results_only import (
        _ResultsOnlyCompatibleModelAlgoFactory,
        _ResultsOnlyCompatibleModellerAlgorithm,
        _ResultsOnlyDataIncompatibleWorkerAlgorithm,
    )
    from bitfount.types import (
        DistributedModelProtocol,
        _DistributedModelTypeOrReference,
    )

    # Check compatible with ResultsOnly
    _algo_factory: _ResultsOnlyCompatibleModelAlgoFactory = ModelEvaluation(
        model=cast(_DistributedModelTypeOrReference, object())
    )
    _modeller_side: _ResultsOnlyCompatibleModellerAlgorithm = _ModellerSide(
        model=cast(DistributedModelProtocol, object())
    )
    _worker_side: _ResultsOnlyDataIncompatibleWorkerAlgorithm = _WorkerSide(
        model=cast(DistributedModelProtocol, object())
    )



"""Tests for the ComputeIntersectionRSA Algorithm."""
from pathlib import Path
import sys
from typing import TYPE_CHECKING, List, Tuple
from unittest.mock import Mock

from _pytest.fixtures import fixture
from _pytest.logging import LogCaptureFixture
from _pytest.tmpdir import TempPathFactory
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
import pandas as pd
import pytest
from pytest_mock import MockerFixture

from bitfount import BitfountHub, _Modeller
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datasources.excel_source import ExcelSource
from bitfount.federated.algorithms.compute_intersection_rsa import (
    ComputeIntersectionRSA,
)
from bitfount.federated.algorithms.compute_intersection_rsa import (
    _ModellerSide as ModellerRSABlind,
)
from bitfount.federated.algorithms.compute_intersection_rsa import (
    _WorkerSide as WorkerRSABlind,
)
from bitfount.federated.algorithms.compute_intersection_rsa import _ModellerSide
from bitfount.federated.algorithms.compute_intersection_rsa import _WorkerSide
from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.exceptions import (
    BlindingError,
    OutOfBoundsError,
    PSIMultiplePodsError,
    PSIMultiTableError,
    UnBlindingError,
)
from bitfount.federated.utils import _ALGORITHMS, hash_set
from bitfount.schemas.utils import bf_dump, bf_load
from tests.utils.helper import (
    create_dataset,
    create_datasource,
    integration_test,
    unit_test,
)


class TestComputeIntersectionRSA:
    """Tests for the ComputeIntersectionRSA algorithm."""

    @fixture
    def rsa_key_pair(self) -> Tuple[RSAPrivateKey, RSAPublicKey]:
        """Fixture for RSA key pair."""
        return _RSAEncryption.generate_key_pair()

    @fixture
    def modeller_rsa_blind(self) -> ModellerRSABlind:
        """Fixture for the Modeller-side RSABLind()."""
        return ComputeIntersectionRSA().modeller()

    @fixture
    def worker_rsa_blind(self) -> WorkerRSABlind:
        """Fixture for the  Worker-side RSABLind()."""
        return ComputeIntersectionRSA().worker()

    @fixture
    def datasource(self) -> BaseSource:
        """Fixture for the DataFrameSource."""
        return create_datasource(classification=False)

    @fixture(scope="class")
    def multi_table_excel_file(self, tmp_path_factory: TempPathFactory) -> Path:
        """Path to multi table excel file."""
        dataframe = create_dataset(classification=False)[:10]
        tmp_path = tmp_path_factory.mktemp("temp_excel")
        filename = tmp_path / "test.xlsx"
        with pd.ExcelWriter(
            filename
        ) as writer:  # type: ignore[abstract] # Reason: This is the documented usage. # noqa: B950
            dataframe.to_excel(writer, index=False, sheet_name="Sheet1")
            dataframe.to_excel(writer, index=False, sheet_name="Sheet2")
        return filename

    @fixture(scope="class")
    def multi_table_excel_source(self, multi_table_excel_file: Path) -> ExcelSource:
        """Multi Table ExcelSource."""
        source = ExcelSource(multi_table_excel_file, sheet_name=["Sheet1", "Sheet2"])
        assert source.multi_table
        return source

    # Modeller-side tests
    @unit_test
    def test_random_factors(
        self, mocker: MockerFixture, modeller_rsa_blind: ModellerRSABlind
    ) -> None:
        """Tests that the random factors generation."""
        modeller_rsa_blind.n = 15
        modeller_rsa_blind.e = 3
        d = 3  # the corresponding private exponent (3*3 = 1 mod (3-1)*(5-1))
        random_numbers = [5, 2, 8, 7, 4, 2]
        mocker.patch(
            "bitfount.federated.algorithms.compute_intersection_rsa.secrets.randbelow",
            side_effect=random_numbers,
        )
        rand_factors = modeller_rsa_blind.generate_random_factors(n_elements=5)
        for (r_inv, r_enc), orig_r in zip(rand_factors, random_numbers[1:]):
            assert r_inv * orig_r % modeller_rsa_blind.n == 1
            assert r_enc**d % modeller_rsa_blind.n == orig_r

    @unit_test
    def test_blind_single_element(self, modeller_rsa_blind: ModellerRSABlind) -> None:
        """Tests that the blinding works as expected."""
        modeller_rsa_blind.n = 15
        modeller_rsa_blind.e = 3
        rf = (1, 7)
        x = 8
        assert modeller_rsa_blind.blind(x, rf) == 11

    @unit_test
    def test_blind_set(self, modeller_rsa_blind: ModellerRSABlind) -> None:
        """Tests that the binding a set works as expected."""
        modeller_rsa_blind.n = 15
        modeller_rsa_blind.e = 3
        modeller_rsa_blind.random_factors = [(1, 7), (6, 11)]
        modeller_rsa_blind.hashed_data = [8, 13]
        assert modeller_rsa_blind.blind_set() == [11, 8]

    @unit_test
    def test_blind_set_raises_error(self, modeller_rsa_blind: ModellerRSABlind) -> None:
        """Tests that the blinding a set raises error.

        An error should be raised if there are not enough random factors.
        """
        modeller_rsa_blind.n = 15
        modeller_rsa_blind.e = 3
        modeller_rsa_blind.random_factors = [(1, 7)]
        modeller_rsa_blind.hashed_data = [8, 13]
        with pytest.raises(BlindingError):
            modeller_rsa_blind.blind_set()

    @unit_test
    def test_unblind_single_element(self, modeller_rsa_blind: ModellerRSABlind) -> None:
        """Tests that the unblinding works as expected."""
        modeller_rsa_blind.n = 15
        modeller_rsa_blind.e = 3
        rf = (1, 7)
        x = 8
        assert modeller_rsa_blind.unblind(x, rf) == 8

    @unit_test
    def test_unblind_set(self, modeller_rsa_blind: ModellerRSABlind) -> None:
        """Tests that the unblinding a set works as expected."""
        modeller_rsa_blind.n = 15
        modeller_rsa_blind.e = 3
        modeller_rsa_blind.random_factors = [(1, 7), (6, 11)]
        modeller_rsa_blind.hashed_data = [8, 13]
        assert modeller_rsa_blind.unblind_set(modeller_rsa_blind.hashed_data) == [8, 3]

    @unit_test
    def test_unblind_set_raises_error(
        self, modeller_rsa_blind: ModellerRSABlind
    ) -> None:
        """Tests that the unblinding a set raises error.

        An error should be raised if there are not enough random factors.
        """
        modeller_rsa_blind.n = 15
        modeller_rsa_blind.e = 3
        modeller_rsa_blind.random_factors = [(1, 7)]
        modeller_rsa_blind.hashed_data = [8, 13]
        with pytest.raises(UnBlindingError):
            modeller_rsa_blind.unblind_set(modeller_rsa_blind.hashed_data)

    @unit_test
    def test_intersect(self, modeller_rsa_blind: ModellerRSABlind) -> None:
        """Tests that intersect returns expected results."""
        Y = pd.DataFrame(
            ["match1", "nomatch1", "match2", "nomatch2", "nomatch3", "match3"],
            columns=["column"],
        )
        B = ["hmatch1", "nomatch1", "hmatch2", "nomatch2", "nomatch3", "hmatch3"]
        bf = ["pod_v1", "pod_v2", "hmatch1", "pod_v3", "hmatch3", "pod_v4", "hmatch2"]
        expected_result = ["match1", "match2", "match3"]
        result = modeller_rsa_blind.intersect(
            Y, B, bf  # type:ignore[arg-type] # Reason: for testing purposes only.
        )
        assert result.values.tolist() == pd.DataFrame(expected_result).values.tolist()

    @integration_test
    def test_blinding_unblinding_works(
        self, modeller_rsa_blind: ModellerRSABlind
    ) -> None:
        """Tests that blinding and unblinding works as expected."""
        modeller_rsa_blind.hashed_data = [1, 2, 3, 4, 5, 6]
        modeller_rsa_blind.n = 15
        modeller_rsa_blind.e = 3
        modeller_rsa_blind.random_factors = modeller_rsa_blind.generate_random_factors(
            6
        )
        blinded_set = modeller_rsa_blind.blind_set()
        assert blinded_set == [
            x * r[1] % modeller_rsa_blind.n
            for x, r in zip(
                modeller_rsa_blind.hashed_data, modeller_rsa_blind.random_factors
            )
        ]
        unblinded_set = modeller_rsa_blind.unblind_set(blinded_set)
        assert unblinded_set == [
            x * r[0] % modeller_rsa_blind.n
            for x, r in zip(blinded_set, modeller_rsa_blind.random_factors)
        ]

    @pytest.mark.parametrize(
        "cols",
        [["TARGET"], ["TARGET", "A", "B", "C"]],
    )
    @unit_test
    def test_modeller_initialise_w_columns(
        self,
        cols: List[str],
        datasource: DataFrameSource,
        modeller_rsa_blind: ModellerRSABlind,
        rsa_key_pair: Tuple[RSAPrivateKey, RSAPublicKey],
    ) -> None:
        """Tests that the modeller is initialised correctly with columns."""
        # Limit the number of records for this unit test
        datasource.dataframe = datasource.dataframe[:10]
        data = datasource.get_data()
        modeller_rsa_blind.initialise(
            public_key=rsa_key_pair[1], datasource=datasource, columns_to_intersect=cols
        )
        assert modeller_rsa_blind.n == rsa_key_pair[1].public_numbers().n
        assert modeller_rsa_blind.e == rsa_key_pair[1].public_numbers().e
        assert modeller_rsa_blind.data.equals(data[cols])
        assert len(modeller_rsa_blind.hashed_data) == data.shape[0]
        assert len(modeller_rsa_blind.random_factors) == data.shape[0]

    @unit_test
    def test_modeller_initialise(
        self,
        datasource: DataFrameSource,
        modeller_rsa_blind: ModellerRSABlind,
        rsa_key_pair: Tuple[RSAPrivateKey, RSAPublicKey],
    ) -> None:
        """Tests that the modeller is initialised correctly."""
        # Limit the number of records for this unit test
        datasource.dataframe = datasource.dataframe[:10]
        data = datasource.get_data()
        modeller_rsa_blind.initialise(public_key=rsa_key_pair[1], datasource=datasource)
        assert modeller_rsa_blind.n == rsa_key_pair[1].public_numbers().n
        assert modeller_rsa_blind.e == rsa_key_pair[1].public_numbers().e
        assert modeller_rsa_blind.data.equals(data)
        assert len(modeller_rsa_blind.hashed_data) == data.shape[0]
        assert len(modeller_rsa_blind.random_factors) == data.shape[0]

    @pytest.mark.parametrize(
        "table,cols",
        [("Sheet1", ["TARGET"]), ("Sheet2", ["TARGET", "A", "B", "C"])],
    )
    @unit_test
    def test_modeller_initialise_multitable_cols(
        self,
        cols: List[str],
        modeller_rsa_blind: ModellerRSABlind,
        multi_table_excel_source: ExcelSource,
        rsa_key_pair: Tuple[RSAPrivateKey, RSAPublicKey],
        table: str,
    ) -> None:
        """Tests that the modeller is initialised for multi-table datasource."""  # noqa: B950
        # Set up modeller datasource
        modeller_datasource = multi_table_excel_source
        modeller_rsa_blind.initialise(
            public_key=rsa_key_pair[1],
            datasource=modeller_datasource,
            columns_to_intersect=cols,
            table=table,
        )
        assert modeller_rsa_blind.n == rsa_key_pair[1].public_numbers().n
        assert modeller_rsa_blind.e == rsa_key_pair[1].public_numbers().e
        assert modeller_rsa_blind.data is not None

    @pytest.mark.parametrize(
        "table",
        ["Sheet1", "Sheet2"],
    )
    @unit_test
    def test_modeller_initialise_multitable(
        self,
        modeller_rsa_blind: ModellerRSABlind,
        multi_table_excel_source: ExcelSource,
        rsa_key_pair: Tuple[RSAPrivateKey, RSAPublicKey],
        table: str,
    ) -> None:
        """Tests that the modeller is initialised for multi-table datasource."""  # noqa: B950
        # Set up modeller datasource
        modeller_datasource = multi_table_excel_source
        modeller_rsa_blind.initialise(
            public_key=rsa_key_pair[1],
            datasource=modeller_datasource,
            table=table,
        )
        assert modeller_rsa_blind.n == rsa_key_pair[1].public_numbers().n
        assert modeller_rsa_blind.e == rsa_key_pair[1].public_numbers().e
        assert modeller_rsa_blind.data is not None

    @unit_test
    def test_modeller_initialise_error_multitable(
        self,
        modeller_rsa_blind: ModellerRSABlind,
        multi_table_excel_source: ExcelSource,
        rsa_key_pair: Tuple[RSAPrivateKey, RSAPublicKey],
    ) -> None:
        """Tests that the modeller `initialise` raises error."""
        # Set up modeller datasource
        with pytest.raises(PSIMultiTableError):
            modeller_rsa_blind.initialise(
                public_key=rsa_key_pair[1],
                datasource=multi_table_excel_source,
            )

    @unit_test
    def test_modeller_hashing(self, modeller_rsa_blind: ModellerRSABlind) -> None:
        """Tests that the hashing function works."""
        dataset = [1, 2, 3, 4, 5, 6]
        hashed_set = hash_set(dataset, hashes.SHA256())
        expected_hash_set = []
        for item in dataset:
            digest = hashes.Hash(hashes.SHA256())
            digest.update(str(item).encode())
            data = digest.finalize()
            expected_hash_set.append(int.from_bytes(data, sys.byteorder))
        for hashed_value, expected_value in zip(hashed_set, expected_hash_set):
            # Check that the output of hashing is an integer.
            assert isinstance(hashed_value, int)
            assert hashed_value == expected_value

    @pytest.mark.parametrize(
        "dataset",
        [
            pd.DataFrame([1, 2, 3, 4, 5, 6], columns=["blah"]),
            pd.DataFrame([[1, 2], [3, 4], [5, 6]], columns=["name", "test"]),
        ],
    )
    @unit_test
    def test_hashing_df(self, dataset: pd.DataFrame) -> None:
        """Tests hashing for a dataframe object."""
        hashed_set = hash_set(dataset, hashes.SHA256())
        expected_hash_set = []
        for _, item in dataset.iterrows():
            item.name = None
            # Check that the 'name' column is not removed, just the index 'name' column
            if "name" in dataset.columns:
                assert dataset["name"] is not None
            digest = hashes.Hash(hashes.SHA256())
            digest.update(str(item).encode())
            data = digest.finalize()
            expected_hash_set.append(int.from_bytes(data, sys.byteorder))
        for hashed_value, expected_value in zip(hashed_set, expected_hash_set):
            # Check that the output of hashing is an integer.
            assert isinstance(hashed_value, int)
            assert hashed_value == expected_value

    @unit_test
    def test_modeller_run(
        self, mocker: MockerFixture, modeller_rsa_blind: ModellerRSABlind
    ) -> None:
        """Test modeller run method."""
        modeller_rsa_blind.data = ["item"]
        mock_unblind = mocker.patch.object(ModellerRSABlind, "unblind_set")
        mock_hash = mocker.patch(
            "bitfount.federated.algorithms.compute_intersection_rsa.hash_set"
        )
        mock_intersect = mocker.patch.object(ModellerRSABlind, "intersect")
        modeller_rsa_blind.run([1], [2])
        mock_unblind.assert_called_once_with([2])
        mock_hash.assert_called_once()
        mock_intersect.assert_called_once()

    # Worker-side tests
    @unit_test
    def test_worker_hasing(self, worker_rsa_blind: WorkerRSABlind) -> None:
        """Tests that the hashing function works."""
        dataset = [1, 2, 3, 4, 5, 6]
        hash_function = hashes.SHA512_256()
        hashed_set = hash_set(dataset, hash_function)
        expected_hash_set = []
        for item in dataset:
            digest = hashes.Hash(hash_function)
            digest.update(str(item).encode())
            data = digest.finalize()
            expected_hash_set.append(int.from_bytes(data, sys.byteorder))
        for hashed_value, expected_value in zip(hashed_set, expected_hash_set):
            # Check that the output of hashing is an integer.
            assert isinstance(hashed_value, int)
            assert hashed_value == expected_value

    @unit_test
    def test_worker_decrypt(self, worker_rsa_blind: WorkerRSABlind) -> None:
        """Tests that the decryption function works."""
        input_data = [1, 2, 3, 4, 5]
        worker_rsa_blind.data = input_data
        worker_rsa_blind.n = 15
        worker_rsa_blind.d = 3
        e = 3  # the corresponding private exponent (3*3 = 1 mod (3-1)*(5-1))
        data = worker_rsa_blind.decrypt_set(worker_rsa_blind.data)
        for item, expected_item in zip(data, input_data):
            assert item == pow(expected_item, worker_rsa_blind.d, worker_rsa_blind.n)
            assert pow(item, e, worker_rsa_blind.n) == expected_item

    @unit_test
    def test_worker_decrypt_raises_error(
        self, worker_rsa_blind: WorkerRSABlind
    ) -> None:
        """Tests that the decryption raises out of bounds error."""
        input_data = [1, 24, 3, 4, 5]
        worker_rsa_blind.data = input_data
        worker_rsa_blind.n = 15
        worker_rsa_blind.d = 3
        with pytest.raises(OutOfBoundsError):
            worker_rsa_blind.decrypt_set(worker_rsa_blind.data)

    @pytest.mark.parametrize(
        "cols",
        [["TARGET"], ["TARGET", "A", "B", "C"]],
    )
    @unit_test
    def test_worker_initialise_cols(
        self,
        cols: List[str],
        datasource: DataFrameSource,
        worker_rsa_blind: WorkerRSABlind,
    ) -> None:
        """Tests that the worker is initialised correctly."""
        # Limit the number of records for this unit test
        datasource.dataframe = datasource.dataframe[:10]
        data = datasource.get_data()
        decrypted_set = worker_rsa_blind.initialise(datasource, cols)
        assert worker_rsa_blind.data.equals(data[cols])
        assert len(decrypted_set) == data.shape[0]

    @unit_test
    def test_worker_initialise(
        self, datasource: DataFrameSource, worker_rsa_blind: WorkerRSABlind
    ) -> None:
        """Tests that the worker is initialised correctly."""
        # Limit the number of records for this unit test
        datasource.dataframe = datasource.dataframe[:10]
        data = datasource.get_data()
        decrypted_set = worker_rsa_blind.initialise(datasource)
        assert worker_rsa_blind.data.equals(data)
        assert len(decrypted_set) == data.shape[0]

    @pytest.mark.parametrize(
        "table,cols",
        [("Sheet1", ["TARGET"]), ("Sheet2", ["TARGET", "A", "B", "C"])],
    )
    @unit_test
    def test_worker_initialise_multitable_columns(
        self,
        cols: List[str],
        multi_table_excel_source: ExcelSource,
        table: str,
        worker_rsa_blind: WorkerRSABlind,
    ) -> None:
        """Tests that the worker is initialised correctly for a multitable."""
        worker_rsa_blind.initialise(
            multi_table_excel_source, columns_to_intersect=cols, table=table
        )
        assert worker_rsa_blind.data is not None

    @pytest.mark.parametrize(
        "table",
        ["Sheet1", "Sheet2"],
    )
    @unit_test
    def test_worker_initialise_multitable(
        self,
        multi_table_excel_source: ExcelSource,
        table: str,
        worker_rsa_blind: WorkerRSABlind,
    ) -> None:
        """Tests that the worker is initialised correctly for a multitable."""
        worker_rsa_blind.initialise(multi_table_excel_source, table=table)
        assert worker_rsa_blind.data is not None

    @unit_test
    def test_worker_initialise_error(
        self,
        caplog: LogCaptureFixture,
        multi_table_excel_source: ExcelSource,
        rsa_key_pair: Tuple[RSAPrivateKey, RSAPublicKey],
        worker_rsa_blind: WorkerRSABlind,
    ) -> None:
        """Tests that the worker `initialise` raises federated error."""
        err_string = "PrivateSetIntersection task on a multitable datasource"

        worker_rsa_blind.initialise(
            public_key=rsa_key_pair[1],
            datasource=multi_table_excel_source,
        )

        assert err_string in caplog.records[0].msg

    @unit_test
    def test_worker_run(
        self, mocker: MockerFixture, worker_rsa_blind: WorkerRSABlind
    ) -> None:
        """Tests the worker.run method."""
        mock_decrypt_set = mocker.patch.object(WorkerRSABlind, "decrypt_set")
        worker_rsa_blind.run([1, 2])
        mock_decrypt_set.assert_called_once_with([1, 2])

    # Integration tests
    @pytest.mark.parametrize(
        "cols",
        [["TARGET"], ["TARGET", "A", "B", "C"]],
    )
    @integration_test
    def test_e2e_rsa_blind_algorithm(
        self,
        cols: List[str],
        modeller_rsa_blind: ModellerRSABlind,
        worker_rsa_blind: WorkerRSABlind,
    ) -> None:
        """Tests RSABLind end-to-end."""
        dataset = create_dataset(classification=False)
        # Set up modeller datasource
        modeller_datasource = DataFrameSource(dataset[:10])
        # Set up worker datasource & columns
        worker_datasource = DataFrameSource(dataset[5:20])

        modeller_blinded_set = modeller_rsa_blind.initialise(
            public_key=worker_rsa_blind.public_key,
            datasource=modeller_datasource,
            columns_to_intersect=cols,
        )
        # Check correct attributes are set during initialisation for modeller
        assert modeller_rsa_blind.n == worker_rsa_blind.n
        assert modeller_rsa_blind.e == worker_rsa_blind.public_key.public_numbers().e
        assert modeller_rsa_blind.data.equals(dataset[:10][cols])
        assert modeller_rsa_blind.hashed_data == hash_set(
            dataset[:10][cols], modeller_rsa_blind.first_hash_function
        )
        assert len(modeller_rsa_blind.random_factors) == 10
        # Check initialisation output is as expected
        assert len(modeller_blinded_set) == 10
        assert [
            item * rf[1] % modeller_rsa_blind.n
            for item, rf in zip(
                modeller_rsa_blind.hashed_data, modeller_rsa_blind.random_factors
            )
        ] == modeller_blinded_set

        worker_decrypted_set = worker_rsa_blind.initialise(worker_datasource, cols)
        # Check correct attributes are set during initialisation for worker
        assert worker_rsa_blind.data.equals(dataset[5:20][cols])
        assert worker_decrypted_set == hash_set(
            worker_rsa_blind.decrypt_set(
                hash_set(worker_rsa_blind.data, worker_rsa_blind.first_hash_function)
            ),
            worker_rsa_blind.second_hash_function,
        )
        # Check worker and modeller run
        worker_decrypted_modeller_set = worker_rsa_blind.run(modeller_blinded_set)
        assert worker_decrypted_modeller_set == worker_rsa_blind.decrypt_set(
            modeller_blinded_set
        )
        # Check that the intersection is output correctly
        # We expect intersection at the last 5 rows in the modeller data
        intersection = modeller_rsa_blind.run(
            worker_decrypted_set, worker_decrypted_modeller_set
        )
        assert (
            intersection.values.tolist()
            == modeller_rsa_blind.data.iloc[5:10].values.tolist()
        )

    # Integration tests
    @pytest.mark.parametrize(
        "table,cols",
        [("Sheet1", ["TARGET"]), ("Sheet2", ["TARGET", "A", "B", "C"])],
    )
    @integration_test
    def test_e2e_rsa_blind_algorithm_multitable(
        self,
        cols: List[str],
        modeller_rsa_blind: ModellerRSABlind,
        multi_table_excel_source: ExcelSource,
        table: str,
        worker_rsa_blind: WorkerRSABlind,
    ) -> None:
        """Tests RSABLind end-to-end."""
        # Set up modeller datasource
        modeller_datasource = multi_table_excel_source
        # Set up worker datasource & columns
        worker_datasource = multi_table_excel_source

        modeller_blinded_set = modeller_rsa_blind.initialise(
            public_key=worker_rsa_blind.public_key,
            datasource=modeller_datasource,
            columns_to_intersect=cols,
            table=table,
        )
        # Check correct attributes are set during initialisation for modeller
        assert modeller_rsa_blind.n == worker_rsa_blind.n
        assert modeller_rsa_blind.e == worker_rsa_blind.public_key.public_numbers().e

        assert len(modeller_rsa_blind.random_factors) == 10
        # Check initialisation output is as expected
        assert len(modeller_blinded_set) == 10
        assert [
            item * rf[1] % modeller_rsa_blind.n
            for item, rf in zip(
                modeller_rsa_blind.hashed_data, modeller_rsa_blind.random_factors
            )
        ] == modeller_blinded_set

        worker_decrypted_set = worker_rsa_blind.initialise(
            worker_datasource, cols, table
        )
        # Check correct attributes are set during initialisation for worker
        assert worker_decrypted_set == hash_set(
            worker_rsa_blind.decrypt_set(
                hash_set(worker_rsa_blind.data, worker_rsa_blind.first_hash_function)
            ),
            worker_rsa_blind.second_hash_function,
        )
        # Check worker and modeller run
        worker_decrypted_modeller_set = worker_rsa_blind.run(modeller_blinded_set)
        assert worker_decrypted_modeller_set == worker_rsa_blind.decrypt_set(
            modeller_blinded_set
        )
        # Check that the intersection is output correctly
        # We expect intersection at the last 5 rows in the modeller data
        intersection = modeller_rsa_blind.run(
            worker_decrypted_set, worker_decrypted_modeller_set
        )

        assert intersection.values.tolist() == modeller_rsa_blind.data.values.tolist()

    @pytest.mark.parametrize(
        "cols",
        [["TARGET"], ["TARGET", "A", "B", "C"]],
    )
    @integration_test
    def test_step_by_step_rsa_blind_algorithm(
        self,
        cols: List[str],
        modeller_rsa_blind: ModellerRSABlind,
        worker_rsa_blind: WorkerRSABlind,
    ) -> None:
        """Test step-by-step ComputeIntersectionRSA Algorithm."""
        worker_rsa_blind = ComputeIntersectionRSA().worker()
        # worker is initialized with the hash functions to use,
        # and the public and private key parameters
        dataset = create_dataset(classification=False)

        # Set up modeller datasource
        modeller_datasource = DataFrameSource(dataset[:10])
        # Set up worker datasource
        worker_datasource = DataFrameSource(dataset[5:20])

        # Algorithm initialisation phases.
        # Modeller initialization:
        (
            modeller_rsa_blind.n,
            modeller_rsa_blind.e,
        ) = modeller_rsa_blind._get_public_key_numbers(worker_rsa_blind.public_key)

        # Load and hash the data with the first hash function for the modeller
        # Take only the 'TARGET' columns from the dataset
        modeller_rsa_blind.data = modeller_datasource.get_data()[cols]
        assert len(modeller_rsa_blind.data) == 10
        modeller_rsa_blind.hashed_data = hash_set(
            modeller_rsa_blind.data, modeller_rsa_blind.first_hash_function
        )
        assert len(modeller_rsa_blind.hashed_data) == 10

        # Check that enough random factors are generated on the modeller side
        modeller_rsa_blind.random_factors = modeller_rsa_blind.generate_random_factors(
            len(modeller_rsa_blind.data)
        )
        assert len(modeller_rsa_blind.random_factors) == 10

        # Check that blind_set works on the modeller side
        blinded_modeller_set = (
            modeller_rsa_blind.blind_set()
        )  # output of modeller_rsa_blind.initialise()
        assert len(blinded_modeller_set) == 10
        assert [
            item * rf[1] % modeller_rsa_blind.n
            for item, rf in zip(
                modeller_rsa_blind.hashed_data, modeller_rsa_blind.random_factors
            )
        ] == blinded_modeller_set

        # Worker Initialization
        # Load and hash the data with the first hash function for the worker
        # Take only the 'TARGET' columns from the dataset
        worker_rsa_blind.data = worker_datasource.get_data()[cols]
        assert len(worker_rsa_blind.data) == 15
        worker_hashed_data = hash_set(
            worker_rsa_blind.data, worker_rsa_blind.first_hash_function
        )
        assert len(worker_hashed_data) == 15

        # Decrypt and hash again the worker set
        decrypted_worker_set = worker_rsa_blind.decrypt_set(worker_hashed_data)
        assert len(decrypted_worker_set) == 15
        assert [
            pow(item, worker_rsa_blind.d, worker_rsa_blind.n)
            for item in worker_hashed_data
        ] == decrypted_worker_set

        # Hash again; this is the set that the modeller receives as the input for psi.
        hashed_decrypted_worker_set = hash_set(
            decrypted_worker_set, worker_rsa_blind.second_hash_function
        )
        assert len(hashed_decrypted_worker_set) == 15

        # Auxiliary checks for intermediary steps

        # Check that both worker and modeller have the same modulus
        assert worker_rsa_blind.n == modeller_rsa_blind.n
        # Check encryption and decryption return the same number
        enc = pow(2, modeller_rsa_blind.e, modeller_rsa_blind.n)
        dec = pow(enc, worker_rsa_blind.d, worker_rsa_blind.n)
        assert dec == 2

        # Check that the random factors can be cancelled out.
        for r_inv, r_enc in modeller_rsa_blind.random_factors:
            assert pow(r_enc, worker_rsa_blind.d, modeller_rsa_blind.n) == pow(
                r_inv, -1, modeller_rsa_blind.n
            )
            assert (
                pow(r_enc, worker_rsa_blind.d, modeller_rsa_blind.n)
                * r_inv
                % modeller_rsa_blind.n
                == 1
            )

        # Check that the hashes for the common elements match
        aux = modeller_rsa_blind.intersect(
            modeller_rsa_blind.data, modeller_rsa_blind.hashed_data, worker_hashed_data
        )
        # We expect intersection at the last 5 rows in the modeller data
        assert aux.values.tolist() == modeller_rsa_blind.data.iloc[5:10].values.tolist()

        # Running the algorithm phases.

        # Worker side:
        # Decrypt modeller blinded set.
        decrypted_modeller_set = worker_rsa_blind.decrypt_set(blinded_modeller_set)
        assert len(decrypted_modeller_set) == 10
        assert [
            pow(item, worker_rsa_blind.d, worker_rsa_blind.n)
            for item in blinded_modeller_set
        ] == decrypted_modeller_set

        # That's all on the worker side, moving on to the modeller_side.

        # First, unblind the modeller set received from the pod
        unblinded_set = modeller_rsa_blind.unblind_set(decrypted_modeller_set)
        assert len(unblinded_set) == 10
        assert [
            item * rf[0] % modeller_rsa_blind.n
            for item, rf in zip(
                decrypted_modeller_set, modeller_rsa_blind.random_factors
            )
        ] == unblinded_set

        # Hash the set with the second hash function
        hashed_unblinded_modeller_set = hash_set(
            unblinded_set, modeller_rsa_blind.second_hash_function
        )
        assert len(hashed_unblinded_modeller_set) == 10
        intersect = modeller_rsa_blind.intersect(
            modeller_rsa_blind.data,
            hashed_unblinded_modeller_set,
            hashed_decrypted_worker_set,
        )

        # Check that the intersection is output correctly
        # We expect intersection at the last 5 rows in the modeller data
        assert (
            intersect.values.tolist()
            == modeller_rsa_blind.data.iloc[5:10].values.tolist()
        )

    # MixIn.intersect() method tests
    @unit_test
    def test_compute_intersection_rsa_execute(
        self, datasource: BaseSource, mock_bitfount_session: Mock, mocker: MockerFixture
    ) -> None:
        """Test ComputeIntersectionRSA.execute()."""
        psi = ComputeIntersectionRSA()
        pod_identifiers = ["username/pod-id"]
        mock_modeller_run_method = mocker.patch.object(_Modeller, "run")
        mock__create_bitfounthub = mocker.patch(
            "bitfount.federated.mixins._create_bitfounthub"
        )
        mock___create_message_service = mocker.patch(
            "bitfount.federated.mixins._create_message_service"
        )
        mock_mock___get_idp_url = mocker.patch("bitfount.federated.mixins._get_idp_url")
        psi.execute(pod_identifiers=pod_identifiers, datasource=datasource)
        mock__create_bitfounthub.assert_called_once()
        mock___create_message_service.assert_called_once()
        mock_mock___get_idp_url.assert_called_once()
        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers=pod_identifiers, require_all_pods=False
        )

    @unit_test
    def test_rsa_blind_execute_error(
        self,
        datasource: BaseSource,
        mock_bitfount_session: Mock,
    ) -> None:
        """Test ComputeIntersectionRSA.execute() raises error."""
        psi = ComputeIntersectionRSA()
        pod_identifiers = ["username/pod-id", "username/pod-id2"]
        with pytest.raises(PSIMultiplePodsError):
            psi.execute(
                pod_identifiers=pod_identifiers,
                datasource=datasource,
            )

    @unit_test
    def test_rsa_blind_execute_columns(
        self,
        datasource: DataFrameSource,
        mock_bitfount_session: Mock,
        mocker: MockerFixture,
    ) -> None:
        """Test ComputeIntersectionRSA.execute() with `columns` argument."""
        pod_identifiers = ["username/pod-id"]
        psi = ComputeIntersectionRSA(
            columns={"self": ["TARGET"], pod_identifiers[0]: ["TARGET"]}
        )
        mock_modeller_run_method = mocker.patch.object(_Modeller, "run")
        psi.execute(pod_identifiers=pod_identifiers, datasource=datasource)
        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers=pod_identifiers, require_all_pods=False
        )


@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for rsa blind algorithm."""

    def test_serialization(self) -> None:
        """Test Marshmallow Serialization for private sql algorithm."""
        algorithm_factory = ComputeIntersectionRSA()
        dumped = bf_dump(algorithm_factory)
        loaded = bf_load(dumped, _ALGORITHMS)
        assert isinstance(loaded, ComputeIntersectionRSA)
        assert algorithm_factory.__dict__ == loaded.__dict__


# Static tests for algorithm-protocol compatibility
if TYPE_CHECKING:
    from typing import cast

    from bitfount.federated.protocols.psi import (
        _PSICompatibleAlgoFactory_,
        _PSICompatibleModeller,
        _PSICompatibleWorker,
    )

    # Check compatible with PrivateSetIntersection
    _algo_factory: _PSICompatibleAlgoFactory_ = ComputeIntersectionRSA()
    _modeller_side: _PSICompatibleModeller = _ModellerSide()
    _worker_side: _PSICompatibleWorker = _WorkerSide(
        hub=cast(BitfountHub, object()),
    )



"""Tests SecureAggregator."""
from typing import Any, Dict, List, Mapping, cast
from unittest.mock import Mock, create_autospec

import numpy as np
import pandas as pd
import pytest
from pytest import fixture

from bitfount.federated.aggregators.aggregator import _ModellerSide, _WorkerSide
from bitfount.federated.aggregators.base import _BaseAggregator
from bitfount.federated.aggregators.base import _registry as aggregator_registry
from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.aggregators.secure import _ModellerSide as SecModellerSide
from bitfount.federated.aggregators.secure import _WorkerSide as SecWorkerSide
from bitfount.federated.exceptions import AggregatorError
from bitfount.federated.secure import SecureShare
from bitfount.federated.shim import BackendTensorShim
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.schemas.utils import bf_dump, bf_load
from bitfount.types import _SerializedWeights, _TensorLike
from tests.bitfount.federated.aggregators.util import assert_equal_weight_dicts
from tests.utils.helper import get_arg_from_args_or_kwargs, unit_test


def asert_vars_equal(agg: Any, dumped: Any) -> None:
    """Helper function for comparing aggregators."""
    for item in vars(agg).keys():
        if item != "_secure_share" and item != "_tensor_shim":
            assert vars(agg)[item] == vars(dumped)[item]


@fixture
def tensor_shim() -> Mock:
    """Returns mock tensor_shim."""
    mock_tensor_shim: Mock = create_autospec(BackendTensorShim)
    mock_tensor_shim.to_list.side_effect = lambda x: x.tolist()
    mock_tensor_shim.to_numpy.side_effect = lambda x: np.asarray(x)
    mock_tensor_shim.to_tensor.side_effect = lambda x, dtype: np.asarray(x, dtype)
    return mock_tensor_shim


@fixture
def secure_share() -> Mock:
    """Returns mock secure share."""
    mock_secure_share: Mock = create_autospec(SecureShare, instance=True)
    return mock_secure_share


@fixture
def mock_mailbox() -> Mock:
    """Mock WorkerMailbox."""
    mailbox: Mock = create_autospec(_WorkerMailbox, instance=True)
    return mailbox


@fixture
def dataframe() -> pd.DataFrame:
    """Returns a dataframe."""
    return pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6], "c": ["7", "8", "9"]})


@unit_test
class TestModellerSide:
    """Test Aggregator ModellerSide."""

    @fixture
    def modeller_side(self, secure_share: Mock, tensor_shim: Mock) -> SecModellerSide:
        """Create ModellerSide for tests."""
        return SecModellerSide(secure_share=secure_share, tensor_shim=tensor_shim)

    def test_run_with_model_parameters(
        self, modeller_side: SecModellerSide, secure_share: Mock
    ) -> None:
        """Test run method with model parameters."""
        parameter_updates: Dict[str, _SerializedWeights] = {
            "user1/pod1": {"hello": [1.0, 1.0, 1.0], "world": [2.0, 2.0, 2.0]},
            "user2/pod2": {"hello": [2.0, 2.0, 2.0], "world": [3.0, 3.0, 3.0]},
        }
        modeller_side.run(algorithm_outputs=parameter_updates)

        # Need to manually check call args due to equality checking numpy arrays
        mock_decoder = secure_share.average_and_decode_state_dicts
        mock_decoder.assert_called_once()

        expected_state_dicts = [
            {k: np.asarray(v) for k, v in i.items()} for i in parameter_updates.values()
        ]
        actual_state_dicts: List[Dict[str, np.ndarray]] = get_arg_from_args_or_kwargs(
            mock_decoder.call_args, 0, "state_dicts"
        )

        assert len(actual_state_dicts) == len(expected_state_dicts)
        for actual, expected in zip(actual_state_dicts, expected_state_dicts):
            assert_equal_weight_dicts(actual, expected)

    def test_run_with_pandas_dataframe(
        self,
        dataframe: pd.DataFrame,
        modeller_side: SecModellerSide,
        secure_share: Mock,
    ) -> None:
        """Test run method with pandas dataframe."""
        dataframes: Dict[str, pd.DataFrame] = {
            "user1/pod1": dataframe,
            "user2/pod2": dataframe,
        }
        secure_share.average_and_decode_state_dicts.return_value = dataframe[
            ["a", "b"]  # c is a string, so it won't be included
        ]
        output = modeller_side.run(algorithm_outputs=dataframes)
        secure_share.average_and_decode_state_dicts.assert_called_once()
        assert isinstance(output, pd.DataFrame)
        assert set(output.columns) == {"a", "b", "c"}

    def test_run_with_numpy_array(
        self,
        dataframe: pd.DataFrame,
        modeller_side: SecModellerSide,
        secure_share: Mock,
    ) -> None:
        """Test run method with numpy array."""
        numpy_arrays: Dict[str, np.ndarray] = {
            "user1/pod1": dataframe["a"].to_numpy(),
            "user2/pod2": dataframe["b"].to_numpy(),
        }
        expected_output = np.mean([dataframe["a"], dataframe["b"]], axis=0)
        secure_share.average_and_decode_state_dicts.return_value = {
            "array": expected_output
        }
        output = modeller_side.run(algorithm_outputs=numpy_arrays)
        secure_share.average_and_decode_state_dicts.assert_called_once()
        assert isinstance(output, np.ndarray)
        np.testing.assert_array_equal(output, expected_output)

    def test_run_with_unrecognised_type(self, modeller_side: SecModellerSide) -> None:
        """Test run method with pandas dataframe."""
        with pytest.raises(AggregatorError):
            # string type not supported
            modeller_side.run({"pod1": "un-recognised type"})  # type: ignore[dict-item] # Reason: Purpose of test # noqa: B950


@unit_test
class TestWorkerSide:
    """Test Aggregator WorkerSide."""

    @fixture
    def worker_side(
        self, mock_mailbox: Mock, secure_share: Mock, tensor_shim: Mock
    ) -> SecWorkerSide:
        """Create WorkerSide for tests."""
        return SecWorkerSide(
            secure_share=secure_share, mailbox=mock_mailbox, tensor_shim=tensor_shim
        )

    async def test_run_with_model_parameters(
        self, secure_share: Mock, worker_side: SecWorkerSide
    ) -> None:
        """Test run method with model parameters."""
        expected_output = {"hello": [1.0, 1.0, 1.0], "world": [2.0, 2.0, 2.0]}
        array_parameter_update: Mapping[str, _TensorLike] = {
            k: cast(_TensorLike, np.asarray(v)) for k, v in expected_output.items()
        }

        # Mock out the secure aggregation
        secure_share.do_secure_aggregation.return_value = array_parameter_update

        output = await worker_side.run(array_parameter_update)
        secure_share.do_secure_aggregation.assert_awaited_once()
        assert output == expected_output

    async def test_run_with_pandas_dataframe(
        self, dataframe: pd.DataFrame, secure_share: Mock, worker_side: SecWorkerSide
    ) -> None:
        """Test run method with pandas dataframe."""
        # Mock out the secure aggregation
        secure_share.do_secure_aggregation.return_value = dataframe[["a", "b"]]

        output = await worker_side.run(dataframe)
        secure_share.do_secure_aggregation.assert_awaited_once()
        # Testing that column c gets added back in (order of columns does not matter)
        pd.testing.assert_frame_equal(
            output.sort_index(axis=1), dataframe.sort_index(axis=1)
        )

    async def test_run_with_numpy_array(
        self, dataframe: pd.DataFrame, secure_share: Mock, worker_side: SecWorkerSide
    ) -> None:
        """Test run method with pandas dataframe."""
        # Mock out the secure aggregation
        expected_output = dataframe["a"].to_numpy()
        secure_share.do_secure_aggregation.return_value = {"array": expected_output}

        output = await worker_side.run(expected_output)
        secure_share.do_secure_aggregation.assert_awaited_once()
        np.testing.assert_equal(output, expected_output)

    async def test_run_with_unrecognised_type(self, worker_side: SecWorkerSide) -> None:
        """Test run method with pandas dataframe."""
        with pytest.raises(AggregatorError):
            # string type is not supported
            await worker_side.run("un-recognised type")  # type: ignore[call-overload] # Reason: Purpose of test # noqa: B950


@unit_test
class TestSecureAggregator:
    """Test Secure Aggregator."""

    def test_modeller(self, secure_share: Mock, tensor_shim: Mock) -> None:
        """Test modeller method."""
        aggregator_factory = SecureAggregator(
            tensor_shim=tensor_shim, secure_share=secure_share
        )
        aggregator = aggregator_factory.modeller()
        for type_ in [
            _BaseAggregator,
            SecModellerSide,
            _ModellerSide,
        ]:
            assert isinstance(aggregator, type_)

    def test_worker(
        self, mock_mailbox: Mock, secure_share: Mock, tensor_shim: Mock
    ) -> None:
        """Test worker method."""
        aggregator_factory = SecureAggregator(
            tensor_shim=tensor_shim, secure_share=secure_share
        )
        aggregator = aggregator_factory.worker(mailbox=mock_mailbox)
        for type_ in [
            _BaseAggregator,
            SecWorkerSide,
            _WorkerSide,
        ]:
            assert isinstance(aggregator, type_)


@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for secure aggregator."""

    def test_dump_load(self) -> None:
        """Test Marshmallow Serialization for secure aggregator."""
        aggregator_factory = SecureAggregator()
        serialized_agg = bf_dump(aggregator_factory)
        deserialized_agg = bf_load(serialized_agg, aggregator_registry)
        asert_vars_equal(deserialized_agg, aggregator_factory)
        asert_vars_equal(
            deserialized_agg._secure_share, aggregator_factory._secure_share
        )



"""Utility modules for making running workers and pods easier.

Modules in this package must never be imported into the core Bitfount package. Anything
defined here is intended for the end user either directly or indirectly e.g. via
scripts, tutorials, etc.
"""

from typing import List

__all__: List[str] = []



"""Generates a schema file from a data file."""
from pathlib import Path
from typing import Any

import fire

from bitfount import config
from bitfount.data.datasources.csv_source import CSVSource
from bitfount.data.schema import BitfountSchema

config._BITFOUNT_CLI_MODE = True


def gen_schema(data_file: str, schema_file: str, **datasource_kwargs: Any) -> None:
    """Generates a schema file from a data file.

    Args:
        data_file: The path to the data file.
        schema_file: The path to save the generated schema to.
        datasource_kwargs: Additional keyword arguments to pass to the schema
                           alongside the data.
    """
    # Create data source
    datasource = CSVSource(Path(data_file).expanduser())

    # Create schema
    if datasource_kwargs:
        print(f"CSVSource kwargs: {datasource_kwargs}")
    schema = BitfountSchema(datasource, **datasource_kwargs)

    # Save schema
    schema.dump(Path(schema_file).expanduser())


def main() -> None:
    """Script entry point."""
    fire.Fire(gen_schema)


if __name__ == "__main__":
    main()



"""Tests model_reference.py module."""
import logging
from pathlib import Path
import sys
from typing import Generator, Type, cast
from unittest.mock import MagicMock

from _pytest.monkeypatch import MonkeyPatch
import pytest
from pytest import fixture
from pytest_mock import MockerFixture

import bitfount
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.federated.model_reference import BitfountModelReference
from bitfount.hub.exceptions import ModelUploadError
from bitfount.models.base_models import MAIN_MODEL_REGISTRY
from bitfount.schemas.utils import bf_dump, bf_load
from bitfount.utils import seed_all
from tests.bitfount import TEST_SECURITY_FILES
from tests.utils.helper import create_datastructure, unit_test

logging.basicConfig(level=logging.INFO, stream=sys.stdout)

PUBLIC_KEY_PATH = TEST_SECURITY_FILES / "test_public.testkey"
PRIVATE_KEY_PATH = TEST_SECURITY_FILES / "test_private.testkey"

seed_all(43)


@unit_test
class TestBitfountModelReference:
    """Tests for BitfountModelReference class."""

    @fixture
    def datastructure(self) -> DataStructure:
        """Fixture for datastructure."""
        return create_datastructure()

    @fixture
    def model_ref(
        self,
        bitfount_model_correct_structure: str,
        datastructure: DataStructure,
        tmp_path: Path,
    ) -> Generator[BitfountModelReference, None, None]:
        """A BitfountModelReference for use in tests with ref as path."""
        model_file = tmp_path / "MyModel.py"
        model_file.touch()
        model_file.write_text(bitfount_model_correct_structure)
        hub_mock = MagicMock()
        hub_mock.send_model.return_value = True
        hub_mock.username = "test_username"
        model_ref = BitfountModelReference(
            username="test",
            datastructure=datastructure,
            schema=BitfountSchema(),
            model_ref=model_file,
            hub=hub_mock,
            hyperparameters={"param1": 1, "param2": 2},
        )
        yield model_ref

    def test_username_is_taken_from_hub_if_none_provided(
        self,
        bitfount_model_correct_structure: str,
        datastructure: DataStructure,
        tmp_path: Path,
    ) -> None:
        """Tests that the username is taken from hub if none provided."""
        model_file = tmp_path / "MyModel.py"
        model_file.touch()
        model_file.write_text(bitfount_model_correct_structure)
        hub_mock = MagicMock()
        hub_mock.username = "test"
        hub_mock.send_model.return_value = True
        model_ref = BitfountModelReference(
            datastructure=datastructure,
            schema=BitfountSchema(),
            model_ref=model_file,
            hub=hub_mock,
            hyperparameters={"param1": 1, "param2": 2},
        )
        assert model_ref.username == "test"

    @pytest.mark.parametrize(
        argnames=("cli_mode", "expected_exc"),
        argvalues=((True, SystemExit), (False, ModelUploadError)),
    )
    def test_get_model_model_sending_unsuccessful(
        self,
        cli_mode: bool,
        expected_exc: Type[BaseException],
        model_ref: BitfountModelReference,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Tests exception raised if method upload fails."""
        # Set _BITFOUNT_CLI_MODE config variable to elicit different error handling
        monkeypatch.setattr(bitfount.config, "_BITFOUNT_CLI_MODE", cli_mode)

        # Make hub.send_model raise a exception
        model_ref.hub = cast(MagicMock, model_ref.hub)
        model_ref.hub.send_model.side_effect = ModelUploadError("ERROR")

        with pytest.raises(expected_exc, match="ERROR"):
            model_ref.get_model()

        model_ref.hub.send_model.assert_called_once()

    def test_get_model_correctly(self, model_ref: BitfountModelReference) -> None:
        """Tests model uploaded and model class retrieved."""
        model_ref.hub = cast(MagicMock, model_ref.hub)
        model_class = model_ref.get_model()
        assert model_class is not None
        assert isinstance(model_class, type)  # i.e. is a class, not an instance
        model_ref.hub.send_model.assert_called_once()

    def test_get_model_as_string(self, model_ref: BitfountModelReference) -> None:
        """Tests model retrieval if model ref is a str."""
        model_ref.hub = cast(MagicMock, model_ref.hub)
        model_ref.hub.get_model.return_value = object
        model_ref.model_ref = "MyModel"
        model_class = model_ref.get_model()
        assert model_class is not None
        assert isinstance(model_class, type)
        model_ref.hub.get_model.assert_called_once()
        model_ref.hub.send_model.assert_not_called()

    def test_get_model_not_recognised(self, model_ref: BitfountModelReference) -> None:
        """Tests exception raised if model ref isn't path or str."""
        model_ref.hub = cast(MagicMock, model_ref.hub)
        model_ref.model_ref = 1123  # type: ignore[assignment]  # Reason: this is the purpose of the test  # noqa: B950
        with pytest.raises(TypeError):
            model_ref.get_model()
        model_ref.hub.send_model.assert_not_called()

    def test_get_model_with_model_that_doesnt_implement_distributed_model_protocol(
        self,
        bitfount_model_incorrect_structure: str,
        mocker: MockerFixture,
        model_ref: BitfountModelReference,
    ) -> None:
        """Tests model not implementing `DistributedModelProtocol` raises TypeError.

        Also ensures that the model doesn't get uploaded to the Hub.
        """
        mock_upload_model_to_hub = mocker.patch.object(
            model_ref, "_upload_model_to_hub"
        )
        assert isinstance(model_ref.model_ref, Path)
        model_ref.model_ref.write_text(bitfount_model_incorrect_structure)
        with pytest.raises(
            TypeError,
            match="Model MyModel does not implement DistributedModelProtocol.",
        ):
            model_ref.get_model()
        mock_upload_model_to_hub.assert_not_called()

    def test_serialize_deserialize(
        self, mocker: MockerFixture, model_ref: BitfountModelReference
    ) -> None:
        """Checks that model reference serialization and deserialization works."""
        # Pre-serialization asserts
        assert model_ref.hub is not None
        assert isinstance(model_ref.model_ref, Path)
        assert model_ref.hyperparameters == {"param1": 1, "param2": 2}

        # Serialization asserts
        dumped = bf_dump(model_ref)
        assert "hub" not in dumped

        with mocker.patch(
            "bitfount.federated.model_reference._default_bitfounthub",
            return_value=model_ref.hub,
        ):
            # Post-deserialization asserts
            loaded = bf_load(dumped, MAIN_MODEL_REGISTRY)
            assert isinstance(loaded.model_ref, str)
            assert loaded.model_ref == "MyModel"
            assert model_ref.username == loaded.username
            assert model_ref.hyperparameters == loaded.hyperparameters

    #
    def test_serialize_deserialize_set_hub_value_works_correctly(
        self, mocker: MockerFixture, model_ref: BitfountModelReference
    ) -> None:
        """Checks that hub attribute is not serialized and is set correctly."""
        # Serialization asserts
        dumped = bf_dump(model_ref)
        assert "hub" not in dumped

        mock_hub = mocker.patch(
            "bitfount.federated.model_reference._default_bitfounthub"
        )

        # Post-deserialization asserts
        bf_load(dumped, MAIN_MODEL_REGISTRY)
        mock_hub.assert_called_once()

    def test_serialization_deserialization_of_model_name_ref(
        self, mocker: MockerFixture, model_ref: BitfountModelReference
    ) -> None:
        """Tests serialization of model when contains model name, not path."""
        model_str = "ModelClassName"
        model_ref.model_ref = model_str  # explicitly set as class name str

        # Serialization asserts - model_ref unchanged
        dumped = bf_dump(model_ref)
        assert dumped["model_ref"] == model_str
        assert "hub" not in dumped

        with mocker.patch("bitfount.federated.model_reference._default_bitfounthub"):

            # Post-deserialization asserts
            loaded = bf_load(dumped, MAIN_MODEL_REGISTRY)
            assert loaded.model_ref == model_str

    @pytest.mark.parametrize(
        "ref_str", ["MyModel.py", "/home/MyModel.py", "/home/MyModel"]
    )
    def test_serialization_of_incorrect_model_ref_fails(
        self, model_ref: BitfountModelReference, ref_str: str
    ) -> None:
        """Tests exception raised if model reference isn't a path.

        Tests with:
            - Basic file name as a string
            - Fuller file path as a string
            - Fuller file path ending in potential class name
        """
        model_ref.model_ref = ref_str  # looks like a path but is a string
        with pytest.raises(
            TypeError,
            match="Unable to serialise model_ref; "
            "expected python file path Path or model name str, "
            f"got <class 'str'> with value {ref_str}",
        ):
            bf_dump(model_ref)

    def test_empty_hyperparameters(
        self, datastructure: DataStructure, mocker: MockerFixture
    ) -> None:
        """Tests an empty dict of hyperparameters is created if none are specified."""
        with mocker.patch("bitfount.federated.model_reference._default_bitfounthub"):
            bfmr = BitfountModelReference(
                model_ref="test",
                datastructure=datastructure,
                schema=BitfountSchema(),
                username="test",
            )
            assert bfmr.hyperparameters == {}
